from collections.abc import Sequence
from typing import Any

import numpy as np
import pytensor
import pytensor.tensor as pt

from pytensor.compile.mode import Mode
from pytensor.tensor.slinalg import solve_discrete_lyapunov

from pymc_extras.statespace.core.statespace import PyMCStateSpace
from pymc_extras.statespace.models.utilities import make_default_coords
from pymc_extras.statespace.utils.constants import (
    ALL_STATE_AUX_DIM,
    ALL_STATE_DIM,
    AR_PARAM_DIM,
    EXOG_STATE_DIM,
    MA_PARAM_DIM,
    OBS_STATE_AUX_DIM,
    OBS_STATE_DIM,
    SHOCK_AUX_DIM,
    SHOCK_DIM,
    TIME_DIM,
)

floatX = pytensor.config.floatX


class BayesianVARMAX(PyMCStateSpace):
    r"""
    Vector AutoRegressive Moving Average with eXogenous Regressors

    The VARMA model is a multivariate extension of the SARIMAX model. Given a set of timeseries :math:`\{x_t\}_{t=0}^T`,
    with :math:`x_t = \begin{bmatrix} x_{1,t} & x_{2,t} & \cdots & x_{k,t} \end{bmatrix}^T`, a VARMA models each series
    as a function of the histories of all series. Specifically, denoting the AR-MA order as (p, q),  a VARMA can be
    written:

    .. math::
        x_t = A_1 x_{t-1} + A_2 x_{t-2} + \cdots + A_p x_{t-p} + B_1 \varepsilon_{t-1} + \cdots
            + B_q \varepsilon_{t-q} + \varepsilon_t

    Where :math:`\varepsilon_t = \begin{bmatrix} \varepsilon_{1,t} & \varepsilon_{2,t} & \cdots &
    \varepsilon_{k,t}\end{bmatrix}^T \sim N(0, \Sigma)` is a vector of i.i.d stochastic innovations or shocks that drive
    intertemporal variation in the data. Matrices :math:`A_i, B_i` are :math:`k \times k` coefficient matrices:

    .. math::
        A_i = \begin{bmatrix} \rho_{1,i,1} & \rho_{1,i,2} & \cdots & \rho_{1,i,k} \\
                              \rho_{2,i,1} & \rho_{2,i,2} & \cdots & \rho_{2,i,k} \\
                              \vdots     &  \vdots    & \cdots & \vdots     \\
                              \rho{k,i,1}  & \rho_{k,i,2} & \cdots & rho_{k,i,k}  \end{bmatrix}

    Internally, this representation is not used. Instead, the vectors :math:`x_t, x_{t-1}, \cdots, x_{t-p},
    \varepsilon_{t-1}, \cdots, \varepsilon_{t-q}` are concatenated into a single column vector of length ``k * (p+q)``,
    while the coefficients matrices are likewise concatenated into a single coefficient matrix, :math:`T`.

    As the dimensionality of the VARMA system increases -- either because there are a large number of timeseries
    included in the analysis, or because the order is large -- the probability of sampling a stationary matrix :math:`T`
    goes to zero. This has two implications for applied work. First, a non-stationary system will exhibit explosive
    behavior, potentially rending impulse response functions and long-term forecasts useless. Secondly, it is not
    possible to do stationary initialization. Stationary initialization significantly speeds up sampling, and should be
    preferred when possible.

    Examples
    --------
    The following code snippet estimates a VARMA(1, 1):

    .. code:: python

        import pymc_extras.statespace as pmss
        import pymc as pm

        # Create VAR Statespace Model
        bvar_mod = pmss.BayesianVARMAX(endog_names=data.columns, order=(2, 0),
                                       stationary_initialization=False, measurement_error=False,
                                       filter_type="standard", verbose=True)

        # Unpack dims and coords
        x0_dims, P0_dims, state_cov_dims, ar_dims = bvar_mod.param_dims.values()
        coords = bvar_mod.coords

        # Estimate PyMC model
        with pm.Model(coords=coords) as var_mod:
            x0 = pm.Normal("x0", dims=x0_dims)
            P0_diag = pm.Gamma("P0_diag", alpha=2, beta=1, size=data.shape[1] * 2, dims=P0_dims[0])
            P0 = pm.Deterministic("P0", pt.diag(P0_diag), dims=P0_dims)

            state_chol, _, _ = pm.LKJCholeskyCov(
                "state_chol", eta=1, n=bvar_mod.k_posdef, sd_dist=pm.Exponential.dist(lam=1)
            )

            ar_params = pm.Normal("ar_params", mu=0, sigma=1, dims=ar_dims)
            state_cov = pm.Deterministic("state_cov", state_chol @ state_chol.T, dims=state_cov_dims)

            bvar_mod.build_statespace_graph(data)
            idata = pm.sample(nuts_sampler="numpyro")
    """

    def __init__(
        self,
        order: tuple[int, int],
        endog_names: list[str] | None = None,
        k_endog: int | None = None,
        exog_state_names: list[str] | dict[str, list[str]] | None = None,
        k_exog: int | dict[str, int] | None = None,
        stationary_initialization: bool = False,
        filter_type: str = "standard",
        measurement_error: bool = False,
        verbose: bool = True,
        mode: str | Mode | None = None,
    ):
        """
        Create a Bayesian VARMAX model.

        Parameters
        ----------
        order: tuple of (int, int)
            Number of autoregressive (AR) and moving average (MA) terms to include in the model. All terms up to the
            specified order are included. For restricted models, set zeros directly on the priors.

        endog_names: list of str, optional
            Names of the endogenous variables being modeled. Used to generate names for the state and shock coords. If
            None, the state names will simply be numbered.

            Exactly one of either ``endog_names`` or ``k_endog`` must be specified.

        exog_state_names : list[str] or dict[str, list[str]], optional
            Names of the exogenous state variables. If a list, all endogenous variables will share the same exogenous
            variables. If a dict, keys should be the names of the endogenous variables, and values should be lists of the
            exogenous variable names for that endogenous variable. Endogenous variables not included in the dict will
            be assumed to have no exogenous variables. If None, no exogenous variables will be included.

        k_exog : int or dict[str, int], optional
            Number of exogenous variables. If an int, all endogenous variables will share the same number of exogenous
            variables. If a dict, keys should be the names of the endogenous variables, and values should be the number of
            exogenous variables for that endogenous variable. Endogenous variables not included in the dict will be
            assumed to have no exogenous variables. If None, no exogenous variables will be included.

        stationary_initialization: bool, default False
            If true, the initial state and initial state covariance will not be assigned priors. Instead, their steady
            state values will be used. If False, the user is responsible for setting priors on the initial state and
            initial covariance.

            ..warning :: This option is very sensitive to the priors placed on the AR and MA parameters. If the model dynamics
                      for a given sample are not stationary, sampling will fail with a "covariance is not positive semi-definite"
                      error.

        filter_type: str, default "standard"
            The type of Kalman Filter to use. Options are "standard", "single", "univariate", "steady_state",
            and "cholesky". See the docs for kalman filters for more details.

        state_structure: str, default "fast"
            How to represent the state-space system. When "interpretable", each element of the state vector will have a
            precise meaning as either lagged data, innovations, or lagged innovations. This comes at the cost of a larger
            state vector, which may hurt performance.

            When "fast", states are combined to minimize the dimension of the state vector, but lags and innovations are
            mixed together as a result. Only the first state (the modeled timeseries) will have an obvious interpretation
            in this case.

        measurement_error: bool, default True
            If true, a measurement error term is added to the model.

        verbose: bool, default True
            If true, a message will be logged to the terminal explaining the variable names, dimensions, and supports.

        mode: str or Mode, optional
            Pytensor compile mode, used in auxiliary sampling methods such as ``sample_conditional_posterior`` and
            ``forecast``. The mode does **not** effect calls to ``pm.sample``.

            Regardless of whether a mode is specified, it can always be overwritten via the ``compile_kwargs`` argument
            to all sampling methods.

        """
        if (endog_names is None) and (k_endog is None):
            raise ValueError("Must specify either endog_names or k_endog")
        if (endog_names is not None) and (k_endog is None):
            k_endog = len(endog_names)
        if (endog_names is None) and (k_endog is not None):
            endog_names = [f"observed_{i}" for i in range(k_endog)]
        if (endog_names is not None) and (k_endog is not None):
            if len(endog_names) != k_endog:
                raise ValueError("Length of provided endog_names does not match provided k_endog")

        if k_exog is not None and not isinstance(k_exog, int | dict):
            raise ValueError("If not None, k_endog must be either an int or a dict")
        if exog_state_names is not None and not isinstance(exog_state_names, list | dict):
            raise ValueError("If not None, exog_state_names must be either a list or a dict")

        if k_exog is not None and exog_state_names is not None:
            if isinstance(k_exog, int) and isinstance(exog_state_names, list):
                if len(exog_state_names) != k_exog:
                    raise ValueError("Length of exog_state_names does not match provided k_exog")
            elif isinstance(k_exog, int) and isinstance(exog_state_names, dict):
                raise ValueError(
                    "If k_exog is an int, exog_state_names must be a list of the same length (or None)"
                )
            elif isinstance(k_exog, dict) and isinstance(exog_state_names, list):
                raise ValueError(
                    "If k_exog is a dict, exog_state_names must be a dict as well (or None)"
                )
            elif isinstance(k_exog, dict) and isinstance(exog_state_names, dict):
                if set(k_exog.keys()) != set(exog_state_names.keys()):
                    raise ValueError("Keys of k_exog and exog_state_names dicts must match")
                if not all(
                    len(names) == k for names, k in zip(exog_state_names.values(), k_exog.values())
                ):
                    raise ValueError(
                        "If both k_endog and exog_state_names are provided, lengths of exog_state_names "
                        "lists must match corresponding values in k_exog"
                    )

        if k_exog is not None and exog_state_names is None:
            if isinstance(k_exog, int):
                exog_state_names = [f"exogenous_{i}" for i in range(k_exog)]
            elif isinstance(k_exog, dict):
                exog_state_names = {
                    name: [f"{name}_exogenous_{i}" for i in range(k)] for name, k in k_exog.items()
                }

        if k_exog is None and exog_state_names is not None:
            if isinstance(exog_state_names, list):
                k_exog = len(exog_state_names)
            elif isinstance(exog_state_names, dict):
                k_exog = {name: len(names) for name, names in exog_state_names.items()}

        # If exog_state_names is a dict but 1) all endog variables are among the keys, and 2) all values are the same
        # then we can drop back to the list case.
        if (
            isinstance(exog_state_names, dict)
            and set(exog_state_names.keys()) == set(endog_names)
            and len({frozenset(val) for val in exog_state_names.values()}) == 1
        ):
            exog_state_names = exog_state_names[endog_names[0]]
            k_exog = len(exog_state_names)

        self.endog_names = list(endog_names)
        self.exog_state_names = exog_state_names

        self.k_exog = k_exog
        self.p, self.q = order
        self.stationary_initialization = stationary_initialization

        k_order = max(self.p, 1) + self.q
        k_states = int(k_endog * k_order)
        k_posdef = k_endog

        super().__init__(
            k_endog,
            k_states,
            k_posdef,
            filter_type,
            verbose=verbose,
            measurement_error=measurement_error,
            mode=mode,
        )

        # Save counts of the number of parameters in each category
        self.param_counts = {
            "x0": k_states * (1 - self.stationary_initialization),
            "P0": k_states**2 * (1 - self.stationary_initialization),
            "AR": k_endog**2 * self.p,
            "MA": k_endog**2 * self.q,
            "state_cov": k_posdef**2,
            "sigma_obs": k_endog * self.measurement_error,
        }

    @property
    def param_names(self):
        names = ["x0", "P0", "ar_params", "ma_params", "state_cov", "sigma_obs"]
        if self.stationary_initialization:
            names.remove("P0")
            names.remove("x0")
        if not self.measurement_error:
            names.remove("sigma_obs")
        if self.p == 0:
            names.remove("ar_params")
        if self.q == 0:
            names.remove("ma_params")

        # Add exogenous regression coefficents rather than remove, since we might have to handle
        # several (if self.exog_state_names is a dict)
        if isinstance(self.exog_state_names, list):
            names.append("beta_exog")
        elif isinstance(self.exog_state_names, dict):
            names.extend([f"beta_{name}" for name in self.exog_state_names.keys()])

        return names

    @property
    def param_info(self) -> dict[str, dict[str, Any]]:
        info = {
            "x0": {
                "shape": (self.k_states,),
                "constraints": None,
            },
            "P0": {
                "shape": (self.k_states, self.k_states),
                "constraints": "Positive Semi-definite",
            },
            "sigma_obs": {
                "shape": (self.k_endog, self.k_endog),
                "constraints": "Positive Semi-definite",
            },
            "state_cov": {
                "shape": (self.k_posdef, self.k_posdef),
                "constraints": "Positive Semi-definite",
            },
            "ar_params": {
                "shape": (self.k_endog, self.p, self.k_endog),
                "constraints": "None",
            },
            "ma_params": {
                "shape": (self.k_endog, self.q, self.k_endog),
                "constraints": "None",
            },
        }

        if isinstance(self.exog_state_names, list):
            k_exog = len(self.exog_state_names)
            info["beta_exog"] = {
                "shape": (self.k_endog, k_exog),
                "constraints": "None",
            }

        elif isinstance(self.exog_state_names, dict):
            for name, exog_names in self.exog_state_names.items():
                k_exog = len(exog_names)
                info[f"beta_{name}"] = {
                    "shape": (k_exog,),
                    "constraints": "None",
                }

        for name in self.param_names:
            info[name]["dims"] = self.param_dims[name]

        return {name: info[name] for name in self.param_names}

    @property
    def data_info(self) -> dict[str, dict[str, Any]]:
        info = None

        if isinstance(self.exog_state_names, list):
            info = {
                "exogenous_data": {
                    "dims": (TIME_DIM, EXOG_STATE_DIM),
                    "shape": (None, self.k_exog),
                }
            }

        elif isinstance(self.exog_state_names, dict):
            info = {
                f"{endog_state}_exogenous_data": {
                    "dims": (TIME_DIM, f"{EXOG_STATE_DIM}_{endog_state}"),
                    "shape": (None, len(exog_names)),
                }
                for endog_state, exog_names in self.exog_state_names.items()
            }

        return info

    @property
    def data_names(self) -> list[str]:
        if isinstance(self.exog_state_names, list):
            return ["exogenous_data"]
        elif isinstance(self.exog_state_names, dict):
            return [f"{endog_state}_exogenous_data" for endog_state in self.exog_state_names.keys()]
        return []

    @property
    def state_names(self):
        state_names = self.endog_names.copy()
        state_names += [
            f"L{i + 1}_{state}" for i in range(self.p - 1) for state in self.endog_names
        ]
        state_names += [
            f"L{i + 1}_{state}_innov" for i in range(self.q) for state in self.endog_names
        ]

        return state_names

    @property
    def observed_states(self):
        return self.endog_names

    @property
    def shock_names(self):
        return self.endog_names

    @property
    def default_priors(self):
        raise NotImplementedError

    @property
    def coords(self) -> dict[str, Sequence]:
        coords = make_default_coords(self)
        if self.p > 0:
            coords.update({AR_PARAM_DIM: list(range(1, self.p + 1))})
        if self.q > 0:
            coords.update({MA_PARAM_DIM: list(range(1, self.q + 1))})

        if isinstance(self.exog_state_names, list):
            coords[EXOG_STATE_DIM] = self.exog_state_names
        elif isinstance(self.exog_state_names, dict):
            for name, exog_names in self.exog_state_names.items():
                coords[f"{EXOG_STATE_DIM}_{name}"] = exog_names

        return coords

    @property
    def param_dims(self):
        coord_map = {
            "x0": (ALL_STATE_DIM,),
            "P0": (ALL_STATE_DIM, ALL_STATE_AUX_DIM),
            "sigma_obs": (OBS_STATE_DIM,),
            "state_cov": (SHOCK_DIM, SHOCK_AUX_DIM),
            "ar_params": (OBS_STATE_DIM, AR_PARAM_DIM, OBS_STATE_AUX_DIM),
            "ma_params": (OBS_STATE_DIM, MA_PARAM_DIM, OBS_STATE_AUX_DIM),
        }

        if not self.measurement_error:
            del coord_map["sigma_obs"]
        if self.p == 0:
            del coord_map["ar_params"]
        if self.q == 0:
            del coord_map["ma_params"]
        if self.stationary_initialization:
            del coord_map["P0"]
            del coord_map["x0"]

        if isinstance(self.exog_state_names, list):
            coord_map["beta_exog"] = (OBS_STATE_DIM, EXOG_STATE_DIM)
        elif isinstance(self.exog_state_names, dict):
            # If each state has its own exogenous variables, each parameter needs it own dim, since we expect the
            # dim labels to all be different (otherwise we'd be in the list case).
            for name in self.exog_state_names.keys():
                coord_map[f"beta_{name}"] = (f"{EXOG_STATE_DIM}_{name}",)

        return coord_map

    def add_default_priors(self):
        raise NotImplementedError

    def make_symbolic_graph(self) -> None:
        # Initialize the matrices
        if not self.stationary_initialization:
            # initial states
            x0 = self.make_and_register_variable("x0", shape=(self.k_states,), dtype=floatX)
            self.ssm["initial_state", :] = x0

            # initial covariance
            P0 = self.make_and_register_variable(
                "P0", shape=(self.k_states, self.k_states), dtype=floatX
            )
            self.ssm["initial_state_cov", :, :] = P0

        # Design matrix is a truncated identity (first k_obs states observed)
        self.ssm[("design", *np.diag_indices(self.k_endog))] = 1

        # Transition matrix has 4 blocks:
        # Upper left: AR coefs (k_obs, k_obs * min(p, 1))
        # Upper right: MA coefs (k_obs, k_obs * q)
        # Lower left: Truncated identity (k_obs * min(p, 1), k_obs * min(p, 1))
        # Lower right: Shifted identity (k_obs * p, k_obs * q)
        self.ssm["transition"] = np.zeros((self.k_states, self.k_states))
        if self.p > 1:
            idx = (
                slice(self.k_endog, self.k_endog * self.p),
                slice(0, self.k_endog * (self.p - 1)),
            )
            self.ssm[("transition", *idx)] = np.eye(self.k_endog * (self.p - 1))

        if self.q > 1:
            idx = (
                slice(-self.k_endog * (self.q - 1), None),
                slice(-self.k_endog * self.q, -self.k_endog),
            )
            self.ssm[("transition", *idx)] = np.eye(self.k_endog * (self.q - 1))

        if self.p > 0:
            ar_param_idx = ("transition", slice(0, self.k_endog), slice(0, self.k_endog * self.p))

            # Register the AR parameter matrix as a (k, p, k), then reshape it and allocate it in the transition matrix
            # This way the user can use 3 dimensions in the prior (clearer?)
            ar_params = self.make_and_register_variable(
                "ar_params", shape=(self.k_endog, self.p, self.k_endog), dtype=floatX
            )

            ar_params = ar_params.reshape((self.k_endog, self.k_endog * self.p))
            self.ssm[ar_param_idx] = ar_params

        # The selection matrix is (k_states, k_obs), with two (k_obs, k_obs) identity
        # matrix blocks inside. One is always on top, the other starts after (k_obs * p) rows
        self.ssm["selection"] = np.zeros((self.k_states, self.k_endog))
        self.ssm["selection", slice(0, self.k_endog), :] = np.eye(self.k_endog)
        if self.q > 0:
            ma_param_idx = (
                "transition",
                slice(0, self.k_endog),
                slice(self.k_endog * max(1, self.p), None),
            )

            # Same as above, register with 3 dimensions then reshape
            ma_params = self.make_and_register_variable(
                "ma_params", shape=(self.k_endog, self.q, self.k_endog), dtype=floatX
            )

            ma_params = ma_params.reshape((self.k_endog, self.k_endog * self.q))
            self.ssm[ma_param_idx] = ma_params

            end = -self.k_endog * (self.q - 1) if self.q > 1 else None
            self.ssm["selection", slice(self.k_endog * -self.q, end), :] = np.eye(self.k_endog)

        if self.measurement_error:
            obs_cov_idx = ("obs_cov", *np.diag_indices(self.k_endog))
            sigma_obs = self.make_and_register_variable(
                "sigma_obs", shape=(self.k_endog,), dtype=floatX
            )
            self.ssm[obs_cov_idx] = sigma_obs

        state_cov = self.make_and_register_variable(
            "state_cov", shape=(self.k_posdef, self.k_posdef), dtype=floatX
        )
        self.ssm["state_cov", :, :] = state_cov

        if self.exog_state_names is not None:
            if isinstance(self.exog_state_names, list):
                beta_exog = self.make_and_register_variable(
                    "beta_exog", shape=(self.k_posdef, self.k_exog), dtype=floatX
                )
                exog_data = self.make_and_register_data(
                    "exogenous_data", shape=(None, self.k_exog), dtype=floatX
                )

                obs_intercept = exog_data @ beta_exog.T

            elif isinstance(self.exog_state_names, dict):
                obs_components = []
                for i, name in enumerate(self.endog_names):
                    if name in self.exog_state_names:
                        k_exog = len(self.exog_state_names[name])
                        beta_exog = self.make_and_register_variable(
                            f"beta_{name}", shape=(k_exog,), dtype=floatX
                        )
                        exog_data = self.make_and_register_data(
                            f"{name}_exogenous_data", shape=(None, k_exog), dtype=floatX
                        )
                        obs_components.append(pt.expand_dims(exog_data @ beta_exog, axis=-1))
                    else:
                        obs_components.append(pt.zeros((1, 1), dtype=floatX))

                # TODO: Replace all of this with pt.concat_with_broadcast once PyMC works with pytensor >= 2.32

                # If there were any zeros, they need to be broadcast against the non-zeros.
                # Core shape is the last dim, the time dim is always broadcast
                non_concat_shape = [1, None]

                # Look for the first non-zero component to get the shape from
                for tensor_inp in obs_components:
                    for i, (bcast, sh) in enumerate(
                        zip(tensor_inp.type.broadcastable, tensor_inp.shape)
                    ):
                        if bcast or i == 1:
                            continue
                        non_concat_shape[i] = sh

                assert non_concat_shape.count(None) == 1

                bcast_tensor_inputs = []
                for tensor_inp in obs_components:
                    non_concat_shape[1] = tensor_inp.shape[1]
                    bcast_tensor_inputs.append(pt.broadcast_to(tensor_inp, non_concat_shape))

                obs_intercept = pt.join(1, *bcast_tensor_inputs)

            else:
                raise NotImplementedError()

            self.ssm["obs_intercept"] = obs_intercept

        if self.stationary_initialization:
            # Solve for matrix quadratic for P0
            T = self.ssm["transition"]
            R = self.ssm["selection"]
            Q = self.ssm["state_cov"]
            c = self.ssm["state_intercept"]

            x0 = pt.linalg.solve(pt.eye(T.shape[0]) - T, c, assume_a="gen", check_finite=False)
            P0 = solve_discrete_lyapunov(
                T,
                pt.linalg.matrix_dot(R, Q, R.T),
                method="direct" if self.k_states < 10 else "bilinear",
            )
            self.ssm["initial_state", :] = x0
            self.ssm["initial_state_cov", :, :] = P0



#   Copyright 2024 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import logging

from collections.abc import Callable
from functools import partial
from typing import Literal
from typing import cast as type_cast

import arviz as az
import numpy as np
import pymc as pm
import pytensor
import pytensor.tensor as pt
import xarray as xr

from better_optimize.constants import minimize_method
from numpy.typing import ArrayLike
from pymc.blocking import DictToArrayBijection
from pymc.model.transform.optimization import freeze_dims_and_data
from pymc.pytensorf import join_nonshared_inputs
from pymc.util import get_default_varnames
from pytensor.graph import vectorize_graph
from pytensor.tensor import TensorVariable
from pytensor.tensor.optimize import minimize
from pytensor.tensor.type import Variable

from pymc_extras.inference.laplace_approx.find_map import (
    _compute_inverse_hessian,
    _make_initial_point,
    find_MAP,
)
from pymc_extras.inference.laplace_approx.scipy_interface import (
    GradientBackend,
    scipy_optimize_funcs_from_loss,
)

_log = logging.getLogger(__name__)


def get_conditional_gaussian_approximation(
    x: TensorVariable,
    Q: TensorVariable | ArrayLike,
    mu: TensorVariable | ArrayLike,
    args: list[TensorVariable] | None = None,
    model: pm.Model | None = None,
    method: minimize_method = "BFGS",
    use_jac: bool = True,
    use_hess: bool = False,
    optimizer_kwargs: dict | None = None,
) -> Callable:
    """
    Returns a function to estimate the a posteriori log probability of a latent Gaussian field x and its mode x0 using the Laplace approximation.

    That is:
    y | x, sigma ~ N(Ax, sigma^2 W)
    x | params ~ N(mu, Q(params)^-1)

    We seek to estimate log(p(x | y, params)):

    log(p(x | y, params)) = log(p(y | x, params)) + log(p(x | params)) + const

    Let f(x) = log(p(y | x, params)). From the definition of our model above, we have log(p(x | params)) = -0.5*(x - mu).T Q (x - mu) + 0.5*logdet(Q).

    This gives log(p(x | y, params)) = f(x) - 0.5*(x - mu).T Q (x - mu) + 0.5*logdet(Q). We will estimate this using the Laplace approximation by Taylor expanding f(x) about the mode.

    Thus:

    1. Maximize log(p(x | y, params)) = f(x) - 0.5*(x - mu).T Q (x - mu) wrt x (note that logdet(Q) does not depend on x) to find the mode x0.

    2. Substitute x0 into the Laplace approximation expanded about the mode: log(p(x | y, params)) ~= -0.5*x.T (-f''(x0) + Q) x + x.T (Q.mu + f'(x0) - f''(x0).x0) + 0.5*logdet(Q).

    Parameters
    ----------
    x: TensorVariable
        The parameter with which to maximize wrt (that is, find the mode in x). In INLA, this is the latent field x~N(mu,Q^-1).
    Q: TensorVariable | ArrayLike
        The precision matrix of the latent field x.
    mu: TensorVariable | ArrayLike
        The mean of the latent field x.
    args: list[TensorVariable]
        Args to supply to the compiled function. That is, (x0, logp) = f(x, *args). If set to None, assumes the model RVs are args.
    model: Model
        PyMC model to use.
    method: minimize_method
        Which minimization algorithm to use.
    use_jac: bool
        If true, the minimizer will compute the gradient of log(p(x | y, params)).
    use_hess: bool
        If true, the minimizer will compute the Hessian log(p(x | y, params)).
    optimizer_kwargs: dict
        Kwargs to pass to scipy.optimize.minimize.

    Returns
    -------
    f: Callable
        A function which accepts a value of x and args and returns [x0, log(p(x | y, params))], where x0 is the mode. x is currently both the point at which to evaluate logp and the initial guess for the minimizer.
    """
    model = pm.modelcontext(model)

    if args is None:
        args = model.continuous_value_vars + model.discrete_value_vars

    # f = log(p(y | x, params))
    f_x = model.logp()
    jac = pytensor.gradient.grad(f_x, x)
    hess = pytensor.gradient.jacobian(jac.flatten(), x)

    # log(p(x | y, params)) only including terms that depend on x for the minimization step (logdet(Q) ignored as it is a constant wrt x)
    log_x_posterior = f_x - 0.5 * (x - mu).T @ Q @ (x - mu)

    # Maximize log(p(x | y, params)) wrt x to find mode x0
    x0, _ = minimize(
        objective=-log_x_posterior,
        x=x,
        method=method,
        jac=use_jac,
        hess=use_hess,
        optimizer_kwargs=optimizer_kwargs,
    )

    # require f'(x0) and f''(x0) for Laplace approx
    jac = pytensor.graph.replace.graph_replace(jac, {x: x0})
    hess = pytensor.graph.replace.graph_replace(hess, {x: x0})

    # Full log(p(x | y, params)) using the Laplace approximation (up to a constant)
    _, logdetQ = pt.nlinalg.slogdet(Q)
    conditional_gaussian_approx = (
        -0.5 * x.T @ (-hess + Q) @ x + x.T @ (Q @ mu + jac - hess @ x0) + 0.5 * logdetQ
    )

    # Currently x is passed both as the query point for f(x, args) = logp(x | y, params) AND as an initial guess for x0. This may cause issues if the query point is
    # far from the mode x0 or in a neighbourhood which results in poor convergence.
    return pytensor.function(args, [x0, conditional_gaussian_approx])


def _unconstrained_vector_to_constrained_rvs(model):
    outputs = get_default_varnames(model.unobserved_value_vars, include_transformed=True)
    constrained_names = [
        x.name for x in get_default_varnames(model.unobserved_value_vars, include_transformed=False)
    ]
    names = [x.name for x in outputs]

    unconstrained_names = [name for name in names if name not in constrained_names]

    new_outputs, unconstrained_vector = join_nonshared_inputs(
        model.initial_point(),
        inputs=model.value_vars,
        outputs=outputs,
    )

    constrained_rvs = [x for x, name in zip(new_outputs, names) if name in constrained_names]
    value_rvs = [x for x in new_outputs if x not in constrained_rvs]

    unconstrained_vector.name = "unconstrained_vector"

    # Redo the names list to ensure it is sorted to match the return order
    names = [*constrained_names, *unconstrained_names]

    return names, constrained_rvs, value_rvs, unconstrained_vector


def model_to_laplace_approx(
    model: pm.Model, unpacked_variable_names: list[str], chains: int = 1, draws: int = 500
):
    initial_point = model.initial_point()
    raveled_vars = DictToArrayBijection.map(initial_point)
    raveled_shape = raveled_vars.data.shape[0]

    # temp_chain and temp_draw are a hack to allow sampling from the Laplace approximation. We only have one mu and cov,
    # so we add batch dims (which correspond to chains and draws). But the names "chain" and "draw" are reserved.
    names, constrained_rvs, value_rvs, unconstrained_vector = (
        _unconstrained_vector_to_constrained_rvs(model)
    )

    coords = model.coords | {
        "temp_chain": np.arange(chains),
        "temp_draw": np.arange(draws),
        "unpacked_variable_names": unpacked_variable_names,
    }

    with pm.Model(coords=coords, model=None) as laplace_model:
        mu = pm.Flat("mean_vector", shape=(raveled_shape,))
        cov = pm.Flat("covariance_matrix", shape=(raveled_shape, raveled_shape))
        laplace_approximation = pm.MvNormal(
            "laplace_approximation",
            mu=mu,
            cov=cov,
            dims=["temp_chain", "temp_draw", "unpacked_variable_names"],
            method="svd",
        )

        cast_to_var = partial(type_cast, Variable)
        batched_rvs = vectorize_graph(
            type_cast(list[Variable], constrained_rvs),
            replace={cast_to_var(unconstrained_vector): cast_to_var(laplace_approximation)},
        )

        for name, batched_rv in zip(names, batched_rvs):
            batch_dims = ("temp_chain", "temp_draw")
            if batched_rv.ndim == 2:
                dims = batch_dims
            elif name in model.named_vars_to_dims:
                dims = (*batch_dims, *model.named_vars_to_dims[name])
            else:
                dims = (*batch_dims, *[f"{name}_dim_{i}" for i in range(batched_rv.ndim - 2)])
                initval = initial_point.get(name, None)
                dim_shapes = initval.shape if initval is not None else batched_rv.type.shape[2:]
                laplace_model.add_coords(
                    {name: np.arange(shape) for name, shape in zip(dims[2:], dim_shapes)}
                )

            pm.Deterministic(name, batched_rv, dims=dims)

    return laplace_model


def unstack_laplace_draws(laplace_data, model, chains=2, draws=500):
    """
    The `model_to_laplace_approx` function returns a model with a single MvNormal distribution, draws from which are
    in the unconstrained variable space. These might be interesting to the user, but since they come back stacked in a
    single vector, it's not easy to work with.

    This function unpacks each component of the vector into its own DataArray, with the appropriate dimensions and
    coordinates, where possible.
    """
    initial_point = DictToArrayBijection.map(model.initial_point())

    cursor = 0
    unstacked_laplace_draws = {}
    coords = model.coords | {"chain": range(chains), "draw": range(draws)}

    # There are corner cases where the value_vars will not have the same dimensions as the random variable (e.g.
    # simplex transform of a Dirichlet). In these cases, we don't try to guess what the labels should be, and just
    # add an arviz-style default dim and label.
    for rv, (name, shape, size, dtype) in zip(model.free_RVs, initial_point.point_map_info):
        rv_dims = []
        for i, dim in enumerate(
            model.named_vars_to_dims.get(rv.name, [f"{name}_dim_{i}" for i in range(len(shape))])
        ):
            if coords.get(dim) and shape[i] == len(coords[dim]):
                rv_dims.append(dim)
            else:
                rv_dims.append(f"{name}_dim_{i}")
                coords[f"{name}_dim_{i}"] = np.arange(shape[i])

        dims = ("chain", "draw", *rv_dims)

        values = (
            laplace_data[..., cursor : cursor + size].reshape((chains, draws, *shape)).astype(dtype)
        )
        unstacked_laplace_draws[name] = xr.DataArray(
            values, dims=dims, coords={dim: list(coords[dim]) for dim in dims}
        )

        cursor += size

    unstacked_laplace_draws = xr.Dataset(unstacked_laplace_draws)

    return unstacked_laplace_draws


def fit_laplace(
    optimize_method: minimize_method | Literal["basinhopping"] = "BFGS",
    *,
    model: pm.Model | None = None,
    use_grad: bool | None = None,
    use_hessp: bool | None = None,
    use_hess: bool | None = None,
    initvals: dict | None = None,
    random_seed: int | np.random.Generator | None = None,
    jitter_rvs: list[pt.TensorVariable] | None = None,
    progressbar: bool = True,
    include_transformed: bool = True,
    gradient_backend: GradientBackend = "pytensor",
    chains: int = 2,
    draws: int = 500,
    optimizer_kwargs: dict | None = None,
    compile_kwargs: dict | None = None,
) -> az.InferenceData:
    """
    Create a Laplace (quadratic) approximation for a posterior distribution.

    This function generates a Laplace approximation for a given posterior distribution using a specified
    number of draws. This is useful for obtaining a parametric approximation to the posterior distribution
    that can be used for further analysis.

    Parameters
    ----------
    model : pm.Model
        The PyMC model to be fit. If None, the current model context is used.
    optimize_method : str
        The optimization method to use. Valid choices are: Nelder-Mead, Powell, CG, BFGS, L-BFGS-B, TNC, SLSQP,
        trust-constr, dogleg, trust-ncg, trust-exact, trust-krylov, and basinhopping.

        See scipy.optimize.minimize documentation for details.
    use_grad : bool | None, optional
        Whether to use gradients in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    use_hessp : bool | None, optional
        Whether to use Hessian-vector products in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    use_hess : bool | None, optional
        Whether to use the Hessian matrix in the optimization. Defaults to None, which determines this automatically based on
        the ``method``.
    initvals : None | dict, optional
        Initial values for the model parameters, as str:ndarray key-value pairs. Paritial initialization is permitted.
         If None, the model's default initial values are used.
    random_seed : None | int | np.random.Generator, optional
        Seed for the random number generator or a numpy Generator for reproducibility
    jitter_rvs : list of TensorVariables, optional
        Variables whose initial values should be jittered. If None, all variables are jittered.
    progressbar : bool, optional
        Whether to display a progress bar during optimization. Defaults to True.
    include_transformed: bool, default True
        Whether to include transformed variables in the output. If True, transformed variables will be included in the
        output InferenceData object. If False, only the original variables will be included.
    gradient_backend: str, default "pytensor"
        The backend to use for gradient computations. Must be one of "pytensor" or "jax".
    chains: int, default: 2
        The number of chain dimensions to sample. Note that this is *not* the number of chains to run in parallel,
        because the Laplace approximation is not an MCMC method. This argument exists to ensure that outputs are
        compatible with the ArviZ library.
    draws: int, default: 500
        The number of samples to draw from the approximated posterior. Totals samples will be chains * draws.
    optimizer_kwargs
        Additional keyword arguments to pass to the ``scipy.optimize`` function being used. Unless
        ``method = "basinhopping"``, ``scipy.optimize.minimize`` will be used. For ``basinhopping``,
        ``scipy.optimize.basinhopping`` will be used. See the documentation of these functions for details.
    compile_kwargs: dict, optional
        Additional keyword arguments to pass to pytensor.function.

    Returns
    -------
    :class:`~arviz.InferenceData`
        An InferenceData object containing the approximated posterior samples.

    Examples
    --------
    >>> from pymc_extras.inference import fit_laplace
    >>> import numpy as np
    >>> import pymc as pm
    >>> import arviz as az
    >>> y = np.array([2642, 3503, 4358]*10)
    >>> with pm.Model() as m:
    >>>     logsigma = pm.Uniform("logsigma", 1, 100)
    >>>     mu = pm.Uniform("mu", -10000, 10000)
    >>>     yobs = pm.Normal("y", mu=mu, sigma=pm.math.exp(logsigma), observed=y)
    >>>     idata = fit_laplace()

    Notes
    -----
    This method of approximation may not be suitable for all types of posterior distributions,
    especially those with significant skewness or multimodality.

    See Also
    --------
    fit : Calling the inference function 'fit' like pmx.fit(method="laplace", model=m)
          will forward the call to 'fit_laplace'.

    """
    compile_kwargs = {} if compile_kwargs is None else compile_kwargs
    optimizer_kwargs = {} if optimizer_kwargs is None else optimizer_kwargs
    model = pm.modelcontext(model) if model is None else model

    idata = find_MAP(
        method=optimize_method,
        model=model,
        use_grad=use_grad,
        use_hessp=use_hessp,
        use_hess=use_hess,
        initvals=initvals,
        random_seed=random_seed,
        jitter_rvs=jitter_rvs,
        progressbar=progressbar,
        include_transformed=include_transformed,
        gradient_backend=gradient_backend,
        compile_kwargs=compile_kwargs,
        compute_hessian=True,
        **optimizer_kwargs,
    )

    unpacked_variable_names = idata.fit["mean_vector"].coords["rows"].values.tolist()

    if "covariance_matrix" not in idata.fit:
        # The user didn't use `use_hess` or `use_hessp` (or an optimization method that returns an inverse Hessian), so
        # we have to go back and compute the Hessian at the MAP point now.
        frozen_model = freeze_dims_and_data(model)
        initial_params = _make_initial_point(frozen_model, initvals, random_seed, jitter_rvs)

        _, f_hessp = scipy_optimize_funcs_from_loss(
            loss=-frozen_model.logp(jacobian=False),
            inputs=frozen_model.continuous_value_vars + frozen_model.discrete_value_vars,
            initial_point_dict=DictToArrayBijection.rmap(initial_params),
            use_grad=False,
            use_hess=False,
            use_hessp=True,
            gradient_backend=gradient_backend,
            compile_kwargs=compile_kwargs,
        )
        H_inv = _compute_inverse_hessian(
            optimizer_result=None,
            optimal_point=idata.fit.mean_vector.values,
            f_fused=None,
            f_hessp=f_hessp,
            use_hess=False,
            method=optimize_method,
        )

        idata.fit["covariance_matrix"] = xr.DataArray(
            H_inv,
            dims=("rows", "columns"),
            coords={"rows": unpacked_variable_names, "columns": unpacked_variable_names},
        )

    with model_to_laplace_approx(model, unpacked_variable_names, chains, draws) as laplace_model:
        new_posterior = (
            pm.sample_posterior_predictive(
                idata.fit.expand_dims(chain=[0], draw=[0]),
                extend_inferencedata=False,
                random_seed=random_seed,
                var_names=[
                    "laplace_approximation",
                    *[x.name for x in laplace_model.deterministics],
                ],
            )
            .posterior_predictive.squeeze(["chain", "draw"])
            .drop_vars(["chain", "draw"])
            .rename({"temp_chain": "chain", "temp_draw": "draw"})
        )

        if include_transformed:
            idata.unconstrained_posterior = unstack_laplace_draws(
                new_posterior.laplace_approximation.values, model, chains=chains, draws=draws
            )

        idata.posterior = new_posterior.drop_vars(
            ["laplace_approximation", "unpacked_variable_names"]
        )

    return idata



import functools as ft
import logging

from collections.abc import Sequence
from itertools import pairwise
from typing import Any

import numpy as np
import xarray as xr

from pytensor import Mode, Variable, config
from pytensor import tensor as pt

from pymc_extras.statespace.core import PyMCStateSpace, PytensorRepresentation
from pymc_extras.statespace.models.utilities import (
    add_tensors_by_dim_labels,
    conform_time_varying_and_time_invariant_matrices,
    join_tensors_by_dim_labels,
    make_default_coords,
)
from pymc_extras.statespace.utils.constants import (
    ALL_STATE_AUX_DIM,
    ALL_STATE_DIM,
    LONG_MATRIX_NAMES,
)

_log = logging.getLogger(__name__)
floatX = config.floatX


class StructuralTimeSeries(PyMCStateSpace):
    r"""
    Structural Time Series Model

    The structural time series model, named by [1] and presented in statespace form in [2], is a framework for
    decomposing a univariate time series into level, trend, seasonal, and cycle components. It also admits the
    possibility of exogenous regressors. Unlike the SARIMAX framework, the time series is not assumed to be stationary.

    Parameters
    ----------
    ssm : PytensorRepresentation
        The state space representation containing system matrices.
    name : str
        Name of the model. If None, defaults to "StructuralTimeSeries".
    state_names : list[str]
        Names of the hidden states in the model.
    observed_state_names : list[str]
        Names of the observed variables.
    data_names : list[str]
        Names of data variables expected by the model.
    shock_names : list[str]
        Names of innovation/shock processes.
    param_names : list[str]
        Names of model parameters.
    exog_names : list[str]
        Names of exogenous variables.
    param_dims : dict[str, tuple[int]]
        Dimension specifications for parameters.
    coords : dict[str, Sequence]
        Coordinate specifications for the model.
    param_info : dict[str, dict[str, Any]]
        Information about parameters including shapes and constraints.
    data_info : dict[str, dict[str, Any]]
        Information about data variables.
    component_info : dict[str, dict[str, Any]]
        Information about model components.
    measurement_error : bool
        Whether the model includes measurement error.
    name_to_variable : dict[str, Variable]
        Mapping from parameter names to PyTensor variables.
    name_to_data : dict[str, Variable] | None, optional
        Mapping from data names to PyTensor variables. Default is None.
    verbose : bool, optional
        Whether to print model information. Default is True.
    filter_type : str, optional
        Type of Kalman filter to use. Default is "standard".
    mode : str | Mode | None, optional
        PyTensor compilation mode. Default is None.

    Notes
    -----
    The structural time series model decomposes a time series into interpretable components:

    .. math::

        y_t = \mu_t + \nu_t + \cdots + \gamma_t + c_t + \xi_t + \varepsilon_t

    Where:
        - :math:`\mu_t` is the level component
        - :math:`\nu_t` is the slope/trend component
        - :math:`\cdots` represents higher-order trend components
        - :math:`\gamma_t` is the seasonal component
        - :math:`c_t` is the cycle component
        - :math:`\xi_t` is the autoregressive component
        - :math:`\varepsilon_t` is the measurement error

    The model is built by combining individual components (e.g., LevelTrendComponent,
    TimeSeasonality, CycleComponent) using the addition operator. Each component
    contributes to the overall state space representation.

    Examples
    --------
    Create a model with trend and seasonal components:

    .. code:: python

        from pymc_extras.statespace import structural as st
        import pymc as pm
        import pytensor.tensor as pt

        trend = st.LevelTrendComponent(order=2 innovations_order=1)
        seasonal = st.TimeSeasonality(season_length=12, innovations=True)
        error = st.MeasurementError()

        ss_mod = (trend + seasonal + error).build()

        with pm.Model(coords=ss_mod.coords) as model:
            P0 = pm.Deterministic('P0', pt.eye(ss_mod.k_states) * 10, dims=ss_mod.param_dims['P0'])

            initial_trend = pm.Normal('initial_trend', sigma=10, dims=ss_mod.param_dims['initial_trend'])
            sigma_trend = pm.HalfNormal('sigma_trend', sigma=1, dims=ss_mod.param_dims['sigma_trend'])

            seasonal_coefs = pm.Normal('params_seasonal', sigma=1, dims=ss_mod.param_dims['params_seasonal'])
            sigma_seasonal = pm.HalfNormal('sigma_seasonal', sigma=1)

            sigma_obs = pm.Exponential('sigma_obs', 1, dims=ss_mod.param_dims['sigma_obs'])

            ss_mod.build_statespace_graph(data)
            idata = pm.sample()

    References
    ----------
    .. [1] Harvey, A. C. (1989). Forecasting, structural time series models and the
           Kalman filter. Cambridge University Press.
    .. [2] Durbin, J., & Koopman, S. J. (2012). Time series analysis by state space
           methods (2nd ed.). Oxford University Press.
    """

    def __init__(
        self,
        ssm: PytensorRepresentation,
        name: str,
        state_names: list[str],
        observed_state_names: list[str],
        data_names: list[str],
        shock_names: list[str],
        param_names: list[str],
        exog_names: list[str],
        param_dims: dict[str, tuple[int]],
        coords: dict[str, Sequence],
        param_info: dict[str, dict[str, Any]],
        data_info: dict[str, dict[str, Any]],
        component_info: dict[str, dict[str, Any]],
        measurement_error: bool,
        name_to_variable: dict[str, Variable],
        name_to_data: dict[str, Variable] | None = None,
        verbose: bool = True,
        filter_type: str = "standard",
        mode: str | Mode | None = None,
    ):
        name = "StructuralTimeSeries" if name is None else name

        self._name = name
        self._observed_state_names = observed_state_names

        k_states, k_posdef, k_endog = ssm.k_states, ssm.k_posdef, ssm.k_endog
        param_names, param_dims, param_info = self._add_inital_state_cov_to_properties(
            param_names, param_dims, param_info, k_states
        )

        self._state_names = self._strip_data_names_if_unambiguous(state_names, k_endog)
        self._data_names = self._strip_data_names_if_unambiguous(data_names, k_endog)
        self._shock_names = self._strip_data_names_if_unambiguous(shock_names, k_endog)
        self._param_names = self._strip_data_names_if_unambiguous(param_names, k_endog)
        self._param_dims = param_dims

        default_coords = make_default_coords(self)
        coords.update(default_coords)

        self._coords = {
            k: self._strip_data_names_if_unambiguous(v, k_endog) for k, v in coords.items()
        }
        self._param_info = param_info.copy()
        self._data_info = data_info.copy()
        self.measurement_error = measurement_error

        super().__init__(
            k_endog,
            k_states,
            max(1, k_posdef),
            filter_type=filter_type,
            verbose=verbose,
            measurement_error=measurement_error,
            mode=mode,
        )
        self.ssm = ssm.copy()

        if k_posdef == 0:
            # If there is no randomness in the model, add dummy matrices to the representation to avoid errors
            # when we go to construct random variables from the matrices
            self.ssm.k_posdef = self.k_posdef
            self.ssm.shapes["state_cov"] = (1, 1, 1)
            self.ssm["state_cov"] = pt.zeros((1, 1, 1))

            self.ssm.shapes["selection"] = (1, self.k_states, 1)
            self.ssm["selection"] = pt.zeros((1, self.k_states, 1))

        self._component_info = component_info.copy()

        self._name_to_variable = name_to_variable.copy()
        self._name_to_data = name_to_data.copy()

        self._exog_names = exog_names.copy()
        self._needs_exog_data = len(exog_names) > 0

        P0 = self.make_and_register_variable("P0", shape=(self.k_states, self.k_states))
        self.ssm["initial_state_cov"] = P0

    def _strip_data_names_if_unambiguous(self, names: list[str], k_endog: int):
        """
        State names from components should always be of the form name[data_name], in the case that the component is
        associated with multiple observed states. Not doing so leads to ambiguity -- we might have two level states,
        but which goes to which observed component? So we set `level[data_1]` and `level[data_2]`.

        In cases where there is only one observed state (when k_endog == 1), we can strip the data part and just use
        the state name. This is a bit cleaner.
        """
        if k_endog == 1:
            [data_name] = self.observed_states
            return [
                name.replace(f"[{data_name}]", "") if isinstance(name, str) else name
                for name in names
            ]

        else:
            return names

    @staticmethod
    def _add_inital_state_cov_to_properties(param_names, param_dims, param_info, k_states):
        param_names += ["P0"]
        param_dims["P0"] = (ALL_STATE_DIM, ALL_STATE_AUX_DIM)
        param_info["P0"] = {
            "shape": (k_states, k_states),
            "constraints": "Positive semi-definite",
            "dims": param_dims["P0"],
        }

        return param_names, param_dims, param_info

    @property
    def param_names(self):
        return self._param_names

    @property
    def data_names(self) -> list[str]:
        return self._data_names

    @property
    def state_names(self):
        return self._state_names

    @property
    def observed_states(self):
        return self._observed_state_names

    @property
    def shock_names(self):
        return self._shock_names

    @property
    def param_dims(self):
        return self._param_dims

    @property
    def coords(self) -> dict[str, Sequence]:
        return self._coords

    @property
    def param_info(self) -> dict[str, dict[str, Any]]:
        return self._param_info

    @property
    def data_info(self) -> dict[str, dict[str, Any]]:
        return self._data_info

    def make_symbolic_graph(self) -> None:
        """
        Assign placeholder pytensor variables among statespace matrices in positions where PyMC variables will go.

        Notes
        -----
        This assignment is handled by the components, so this function is implemented only to avoid the
        NotImplementedError raised by the base class.
        """

        pass

    def _state_slices_from_info(self):
        info = self._component_info.copy()
        comp_states = np.cumsum([0] + [info["k_states"] for info in info.values()])
        state_slices = [slice(i, j) for i, j in pairwise(comp_states)]

        return state_slices

    def _hidden_states_from_data(self, data):
        state_slices = self._state_slices_from_info()
        info = self._component_info
        names = info.keys()
        result = []

        for i, (name, s) in enumerate(zip(names, state_slices)):
            obs_idx = info[name]["obs_state_idx"]

            if obs_idx is None:
                continue

            X = data[..., s]

            if info[name]["combine_hidden_states"]:
                sum_idx_joined = np.flatnonzero(obs_idx)
                sum_idx_split = np.split(sum_idx_joined, info[name]["k_endog"])
                for sum_idx in sum_idx_split:
                    result.append(X[..., sum_idx].sum(axis=-1)[..., None])
            else:
                n_components = len(self.state_names[s])
                for j in range(n_components):
                    result.append(X[..., j, None])

        return np.concatenate(result, axis=-1)

    def _get_subcomponent_names(self):
        state_slices = self._state_slices_from_info()
        info = self._component_info
        names = info.keys()
        result = []

        for i, (name, s) in enumerate(zip(names, state_slices)):
            if info[name]["combine_hidden_states"]:
                if self.k_endog == 1:
                    result.append(name)
                else:
                    # If there are multiple observed states, we will combine per hidden state, preserving the
                    # observed state names. Note this happens even if this *component* has only 1 state for consistency,
                    # as long as the statespace model has multiple observed states.
                    result.extend(
                        [f"{name}[{obs_name}]" for obs_name in info[name]["observed_state_names"]]
                    )
            else:
                comp_names = self.state_names[s]
                result.extend([f"{name}[{comp_name}]" for comp_name in comp_names])
        return result

    def extract_components_from_idata(self, idata: xr.Dataset) -> xr.Dataset:
        r"""
        Extract interpretable hidden states from an InferenceData returned by a PyMCStateSpace sampling method

        Parameters
        ----------
        idata: Dataset
            A Dataset object, returned by a PyMCStateSpace sampling method

        Returns
        -------
        idata: Dataset
            A Dataset object with hidden states transformed to represent only the "interpretable" subcomponents
            of the structural model.

        Notes
        -----
        In general, a structural statespace model can be represented as:

        .. math::
            y_t = \mu_t + \nu_t + \cdots + \gamma_t + c_t + \xi_t + \epsilon_t \tag{1}

        Where:

            - :math:`\mu_t` is the level of the data at time t
            - :math:`\nu_t` is the slope of the data at time t
            - :math:`\cdots` are higher time derivatives of the position (acceleration, jerk, etc) at time t
            - :math:`\gamma_t` is the seasonal component at time t
            - :math:`c_t` is the cycle component at time t
            - :math:`\xi_t` is the autoregressive error at time t
            - :math:`\varepsilon_t` is the measurement error at time t

        In state space form, some or all of these components are represented as linear combinations of other
        subcomponents, making interpretation of the outputs of the outputs difficult. The purpose of this function is
        to take the expended statespace representation and return a "reduced form" of only the components shown in
        equation (1).
        """

        def _extract_and_transform_variable(idata, new_state_names):
            *_, time_dim, state_dim = idata.dims
            state_func = ft.partial(self._hidden_states_from_data)
            new_idata = xr.apply_ufunc(
                state_func,
                idata,
                input_core_dims=[[time_dim, state_dim]],
                output_core_dims=[[time_dim, state_dim]],
                exclude_dims={state_dim},
            )
            new_idata.coords.update({state_dim: new_state_names})
            return new_idata

        var_names = list(idata.data_vars.keys())
        is_latent = [idata[name].shape[-1] == self.k_states for name in var_names]
        new_state_names = self._get_subcomponent_names()

        latent_names = [name for latent, name in zip(is_latent, var_names) if latent]
        dropped_vars = set(var_names) - set(latent_names)
        if len(dropped_vars) > 0:
            _log.warning(
                f"Variables {', '.join(dropped_vars)} do not contain all hidden states (their last dimension "
                f"is not {self.k_states}). They will not be present in the modified idata."
            )
        if len(dropped_vars) == len(var_names):
            raise ValueError(
                "Provided idata had no variables with all hidden states; cannot extract components."
            )

        idata_new = xr.Dataset(
            {
                name: _extract_and_transform_variable(idata[name], new_state_names)
                for name in latent_names
            }
        )
        return idata_new


class Component:
    r"""
    Base class for a component of a structural timeseries model.

    This base class contains a subset of the class attributes of the PyMCStateSpace class, and none of the class
    methods. The purpose of a component is to allow the partial definition of a structural model. Components are
    assembled into a full model by the StructuralTimeSeries class.

    Parameters
    ----------
    name : str
        The name of the component.
    k_endog : int
        Number of endogenous (observed) variables being modeled.
    k_states : int
        Number of hidden states in the component model.
    k_posdef : int
        Rank of the state covariance matrix, or the number of sources of innovations
        in the component model.
    state_names : list[str] | None, optional
        Names of the hidden states. If None, defaults to empty list.
    observed_state_names : list[str] | None, optional
        Names of the observed states associated with this component. Must have the same
        length as k_endog. If None, defaults to empty list.
    data_names : list[str] | None, optional
        Names of data variables expected by the component. If None, defaults to empty list.
    shock_names : list[str] | None, optional
        Names of innovation/shock processes. If None, defaults to empty list.
    param_names : list[str] | None, optional
        Names of component parameters. If None, defaults to empty list.
    exog_names : list[str] | None, optional
        Names of exogenous variables. If None, defaults to empty list.
    representation : PytensorRepresentation | None, optional
        Pre-existing state space representation. If None, creates a new one.
    measurement_error : bool, optional
        Whether the component includes measurement error. Default is False.
    combine_hidden_states : bool, optional
        Whether to combine hidden states when extracting from data. Should be True for
        components where individual states have no interpretation (e.g., seasonal,
        autoregressive). Default is True.
    component_from_sum : bool, optional
        Whether this component is created from combining other components. Default is False.
    obs_state_idxs : np.ndarray | None, optional
        Indices indicating which states contribute to observed variables. If None,
        defaults to None.
    share_states : bool, optional
        Whether states are shared across multiple endogenous variables in multivariate
        models. When True, the same latent states affect all observed variables.
        Default is False.

    Examples
    --------
    Create a simple trend component:

    .. code:: python

        from pymc_extras.statespace import structural as st

        trend = st.LevelTrendComponent(order=2, innovations_order=1)
        seasonal = st.TimeSeasonality(season_length=12, innovations=True)
        model = (trend + seasonal).build()

        print(f"Model has {model.k_states} states and {model.k_posdef} innovations")

    See Also
    --------
    StructuralTimeSeries : The complete model class that combines components.
    LevelTrendComponent : Component for modeling level and trend.
    TimeSeasonality : Component for seasonal effects.
    CycleComponent : Component for cyclical effects.
    RegressionComponent : Component for regression effects.
    """

    def __init__(
        self,
        name,
        k_endog,
        k_states,
        k_posdef,
        state_names=None,
        observed_state_names=None,
        data_names=None,
        shock_names=None,
        param_names=None,
        exog_names=None,
        representation: PytensorRepresentation | None = None,
        measurement_error=False,
        combine_hidden_states=True,
        component_from_sum=False,
        obs_state_idxs=None,
        share_states: bool = False,
    ):
        self.name = name
        self.k_endog = k_endog
        self.k_states = k_states
        self.share_states = share_states
        self.k_posdef = k_posdef
        self.measurement_error = measurement_error

        self.state_names = list(state_names) if state_names is not None else []
        self.observed_state_names = (
            list(observed_state_names) if observed_state_names is not None else []
        )
        self.data_names = list(data_names) if data_names is not None else []
        self.shock_names = list(shock_names) if shock_names is not None else []
        self.param_names = list(param_names) if param_names is not None else []
        self.exog_names = list(exog_names) if exog_names is not None else []

        self.needs_exog_data = len(self.exog_names) > 0
        self.coords = {}
        self.param_dims = {}

        self.param_info = {}
        self.data_info = {}

        self.param_counts = {}

        if representation is None:
            self.ssm = PytensorRepresentation(k_endog=k_endog, k_states=k_states, k_posdef=k_posdef)
        else:
            self.ssm = representation

        self._name_to_variable = {}
        self._name_to_data = {}

        if not component_from_sum:
            self.populate_component_properties()
            self.make_symbolic_graph()

        self._component_info = {
            self.name: {
                "k_states": self.k_states,
                "k_endog": self.k_endog,
                "k_posdef": self.k_posdef,
                "observed_state_names": self.observed_state_names,
                "combine_hidden_states": combine_hidden_states,
                "obs_state_idx": obs_state_idxs,
                "share_states": self.share_states,
            }
        }

    def make_and_register_variable(self, name, shape, dtype=floatX) -> Variable:
        r"""
        Helper function to create a pytensor symbolic variable and register it in the _name_to_variable dictionary

        Parameters
        ----------
        name : str
            The name of the placeholder variable. Must be the name of a model parameter.
        shape : int or tuple of int
            Shape of the parameter
        dtype : str, default pytensor.config.floatX
            dtype of the parameter

        Notes
        -----
        Symbolic pytensor variables are used in the ``make_symbolic_graph`` method as placeholders for PyMC random
        variables. The change is made in the ``_insert_random_variables`` method via ``pytensor.graph_replace``. To
        make the change, a dictionary mapping pytensor variables to PyMC random variables needs to be constructed.

        The purpose of this method is to:
            1.  Create the placeholder symbolic variables
            2.  Register the placeholder variable in the ``_name_to_variable`` dictionary

        The shape provided here will define the shape of the prior that will need to be provided by the user.

        An error is raised if the provided name has already been registered, or if the name is not present in the
        ``param_names`` property.
        """
        if name not in self.param_names:
            raise ValueError(
                f"{name} is not a model parameter. All placeholder variables should correspond to model "
                f"parameters."
            )

        if name in self._name_to_variable.keys():
            raise ValueError(
                f"{name} is already a registered placeholder variable with shape "
                f"{self._name_to_variable[name].type.shape}"
            )

        placeholder = pt.tensor(name, shape=shape, dtype=dtype)
        self._name_to_variable[name] = placeholder
        return placeholder

    def make_and_register_data(self, name, shape, dtype=floatX) -> Variable:
        r"""
        Helper function to create a pytensor symbolic variable and register it in the _name_to_data dictionary

        Parameters
        ----------
        name : str
            The name of the placeholder data. Must be the name of an expected data variable.
        shape : int or tuple of int
            Shape of the parameter
        dtype : str, default pytensor.config.floatX
            dtype of the parameter

        Notes
        -----
        See docstring for make_and_register_variable for more details. This function is similar, but handles data
        inputs instead of model parameters.

        An error is raised if the provided name has already been registered, or if the name is not present in the
        ``data_names`` property.
        """
        if name not in self.data_names:
            raise ValueError(
                f"{name} is not a model parameter. All placeholder variables should correspond to model "
                f"parameters."
            )

        if name in self._name_to_data.keys():
            raise ValueError(
                f"{name} is already a registered placeholder variable with shape "
                f"{self._name_to_data[name].type.shape}"
            )

        placeholder = pt.tensor(name, shape=shape, dtype=dtype)
        self._name_to_data[name] = placeholder
        return placeholder

    def make_symbolic_graph(self) -> None:
        raise NotImplementedError

    def populate_component_properties(self):
        raise NotImplementedError

    def _get_combined_shapes(self, other):
        k_states = self.k_states + other.k_states
        k_posdef = self.k_posdef + other.k_posdef

        # To count endog states, we have to count unique names between the two components.
        combined_states = self._combine_property(
            other, "observed_state_names", allow_duplicates=False
        )
        k_endog = len(combined_states)

        return k_states, k_posdef, k_endog

    def _combine_statespace_representations(self, other):
        def make_slice(name, x, o_x):
            ndim = max(x.ndim, o_x.ndim)
            return (name,) + (slice(None, None, None),) * ndim

        k_states, k_posdef, k_endog = self._get_combined_shapes(other)

        self_matrices = [self.ssm[name] for name in LONG_MATRIX_NAMES]
        other_matrices = [other.ssm[name] for name in LONG_MATRIX_NAMES]

        self_observed_states = self.observed_state_names
        other_observed_states = other.observed_state_names

        x0, P0, c, d, T, Z, R, H, Q = (
            self.ssm[make_slice(name, x, o_x)]
            for name, x, o_x in zip(LONG_MATRIX_NAMES, self_matrices, other_matrices)
        )
        o_x0, o_P0, o_c, o_d, o_T, o_Z, o_R, o_H, o_Q = (
            other.ssm[make_slice(name, x, o_x)]
            for name, x, o_x in zip(LONG_MATRIX_NAMES, self_matrices, other_matrices)
        )

        initial_state = pt.concatenate(conform_time_varying_and_time_invariant_matrices(x0, o_x0))
        initial_state.name = x0.name

        initial_state_cov = pt.linalg.block_diag(P0, o_P0)
        initial_state_cov.name = P0.name

        state_intercept = pt.concatenate(conform_time_varying_and_time_invariant_matrices(c, o_c))
        state_intercept.name = c.name

        obs_intercept = add_tensors_by_dim_labels(
            d, o_d, labels=self_observed_states, other_labels=other_observed_states, labeled_axis=-1
        )
        obs_intercept.name = d.name

        transition = pt.linalg.block_diag(T, o_T)
        transition = pt.specify_shape(
            transition,
            shape=[
                sum(shapes) if not any([s is None for s in shapes]) else None
                for shapes in zip(*[T.type.shape, o_T.type.shape])
            ],
        )
        transition.name = T.name

        design = join_tensors_by_dim_labels(
            *conform_time_varying_and_time_invariant_matrices(Z, o_Z),
            labels=self_observed_states,
            other_labels=other_observed_states,
            labeled_axis=-2,
            join_axis=-1,
        )
        design.name = Z.name

        selection = pt.linalg.block_diag(R, o_R)
        selection = pt.specify_shape(
            selection,
            shape=[
                sum(shapes) if not any([s is None for s in shapes]) else None
                for shapes in zip(*[R.type.shape, o_R.type.shape])
            ],
        )
        selection.name = R.name

        obs_cov = add_tensors_by_dim_labels(
            H,
            o_H,
            labels=self_observed_states,
            other_labels=other_observed_states,
            labeled_axis=(-1, -2),
        )
        obs_cov.name = H.name

        state_cov = pt.linalg.block_diag(Q, o_Q)
        state_cov.name = Q.name

        new_ssm = PytensorRepresentation(
            k_endog=k_endog,
            k_states=k_states,
            k_posdef=k_posdef,
            initial_state=initial_state,
            initial_state_cov=initial_state_cov,
            state_intercept=state_intercept,
            obs_intercept=obs_intercept,
            transition=transition,
            design=design,
            selection=selection,
            obs_cov=obs_cov,
            state_cov=state_cov,
        )

        return new_ssm

    def _combine_property(self, other, name, allow_duplicates=True):
        self_prop = getattr(self, name)
        other_prop = getattr(other, name)

        if not isinstance(self_prop, type(other_prop)):
            raise TypeError(
                f"Property {name} of {self} and {other} are not the same and cannot be combined. Found "
                f"{type(self_prop)} for {self} and {type(other_prop)} for {other}'"
            )

        if not isinstance(self_prop, list | dict):
            raise TypeError(
                f"All component properties are expected to be lists or dicts, but found {type(self_prop)}"
                f"for property {name} of {self} and {type(other_prop)} for {other}'"
            )

        if isinstance(self_prop, list) and allow_duplicates:
            return self_prop + other_prop
        elif isinstance(self_prop, list) and not allow_duplicates:
            return self_prop + [x for x in other_prop if x not in self_prop]
        elif isinstance(self_prop, dict):
            new_prop = self_prop.copy()
            new_prop.update(other_prop)
            return new_prop

    def _combine_component_info(self, other):
        combined_info = {}
        for key, value in self._component_info.items():
            if not key.startswith("StateSpace"):
                if key in combined_info.keys():
                    raise ValueError(f"Found duplicate component named {key}")
                combined_info[key] = value

        for key, value in other._component_info.items():
            if not key.startswith("StateSpace"):
                if key in combined_info.keys():
                    raise ValueError(f"Found duplicate component named {key}")
                combined_info[key] = value

        return combined_info

    def _make_combined_name(self):
        components = self._component_info.keys()
        name = f"StateSpace[{', '.join(components)}]"
        return name

    def __add__(self, other):
        state_names = self._combine_property(other, "state_names")
        data_names = self._combine_property(other, "data_names")
        observed_state_names = self._combine_property(
            other, "observed_state_names", allow_duplicates=False
        )

        param_names = self._combine_property(other, "param_names")
        shock_names = self._combine_property(other, "shock_names")
        param_info = self._combine_property(other, "param_info")
        data_info = self._combine_property(other, "data_info")
        param_dims = self._combine_property(other, "param_dims")
        coords = self._combine_property(other, "coords")
        exog_names = self._combine_property(other, "exog_names")

        _name_to_variable = self._combine_property(other, "_name_to_variable")
        _name_to_data = self._combine_property(other, "_name_to_data")

        measurement_error = any([self.measurement_error, other.measurement_error])

        k_states, k_posdef, k_endog = self._get_combined_shapes(other)

        ssm = self._combine_statespace_representations(other)

        new_comp = Component(
            name="",
            k_endog=k_endog,
            k_states=k_states,
            k_posdef=k_posdef,
            observed_state_names=observed_state_names,
            measurement_error=measurement_error,
            representation=ssm,
            component_from_sum=True,
        )
        new_comp._component_info = self._combine_component_info(other)
        new_comp.name = new_comp._make_combined_name()

        names_and_props = [
            ("state_names", state_names),
            ("observed_state_names", observed_state_names),
            ("data_names", data_names),
            ("param_names", param_names),
            ("shock_names", shock_names),
            ("param_dims", param_dims),
            ("coords", coords),
            ("param_dims", param_dims),
            ("param_info", param_info),
            ("data_info", data_info),
            ("exog_names", exog_names),
            ("_name_to_variable", _name_to_variable),
            ("_name_to_data", _name_to_data),
        ]

        for prop, value in names_and_props:
            setattr(new_comp, prop, value)

        return new_comp

    def build(
        self, name=None, filter_type="standard", verbose=True, mode: str | Mode | None = None
    ):
        """
        Build a StructuralTimeSeries statespace model from the current component(s)

        Parameters
        ----------
        name: str, optional
            Name of the exogenous data being modeled. Default is "data"

        filter_type : str, optional
            The type of Kalman filter to use. Valid options are "standard", "univariate", "single", "cholesky", and
            "steady_state". For more information, see the docs for each filter. Default is "standard".

        verbose : bool, optional
            If True, displays information about the initialized model. Defaults to True.

        mode: str or Mode, optional
            Pytensor compile mode, used in auxiliary sampling methods such as ``sample_conditional_posterior`` and
            ``forecast``. The mode does **not** effect calls to ``pm.sample``.

            Regardless of whether a mode is specified, it can always be overwritten via the ``compile_kwargs`` argument
            to all sampling methods.

        Returns
        -------
        PyMCStateSpace
            An initialized instance of a PyMCStateSpace, constructed using the system matrices contained in the
            components.
        """

        return StructuralTimeSeries(
            self.ssm,
            name=name,
            state_names=self.state_names,
            observed_state_names=self.observed_state_names,
            data_names=self.data_names,
            shock_names=self.shock_names,
            param_names=self.param_names,
            param_dims=self.param_dims,
            coords=self.coords,
            param_info=self.param_info,
            data_info=self.data_info,
            component_info=self._component_info,
            measurement_error=self.measurement_error,
            exog_names=self.exog_names,
            name_to_variable=self._name_to_variable,
            name_to_data=self._name_to_data,
            filter_type=filter_type,
            verbose=verbose,
            mode=mode,
        )



import warnings

import numpy as np
import pymc as pm
import pytensor
import pytensor.tensor as pt
import pytest

from pymc.model.transform.optimization import freeze_dims_and_data
from pymc.testing import mock_sample_setup_and_teardown

from pymc_extras.statespace.utils.constants import (
    FILTER_OUTPUT_NAMES,
    MATRIX_NAMES,
    SMOOTHER_OUTPUT_NAMES,
)
from tests.statespace.core.test_statespace import (  # pylint: disable=unused-import
    exog_ss_mod,
    ss_mod,
)
from tests.statespace.shared_fixtures import (  # pylint: disable=unused-import
    rng,
)
from tests.statespace.test_utilities import load_nile_test_data

pytest.importorskip("jax")
pytest.importorskip("numpyro")


floatX = pytensor.config.floatX
nile = load_nile_test_data()
ALL_SAMPLE_OUTPUTS = MATRIX_NAMES + FILTER_OUTPUT_NAMES + SMOOTHER_OUTPUT_NAMES

mock_pymc_sample = pytest.fixture(scope="session")(mock_sample_setup_and_teardown)


@pytest.fixture(scope="session")
def pymc_mod(ss_mod):
    with pm.Model(coords=ss_mod.coords) as pymc_mod:
        rho = pm.Beta("rho", 1, 1)
        zeta = pm.Deterministic("zeta", 1 - rho)

        ss_mod.build_statespace_graph(data=nile, save_kalman_filter_outputs_in_idata=True)
        names = ["x0", "P0", "c", "d", "T", "Z", "R", "H", "Q"]
        for name, matrix in zip(names, ss_mod.unpack_statespace()):
            pm.Deterministic(name, matrix)

    return pymc_mod


@pytest.fixture(scope="session")
def exog_pymc_mod(exog_ss_mod, rng):
    y = rng.normal(size=(100, 1)).astype(floatX)
    X = rng.normal(size=(100, 3)).astype(floatX)

    with pm.Model(coords=exog_ss_mod.coords) as m:
        exog_data = pm.Data("data_exog", X)
        initial_trend = pm.Normal("initial_trend", dims=["trend_state"])
        P0_sigma = pm.Exponential("P0_sigma", 1)
        P0 = pm.Deterministic(
            "P0", pt.eye(exog_ss_mod.k_states) * P0_sigma, dims=["state", "state_aux"]
        )
        beta_exog = pm.Normal("beta_exog", dims=["exog_state"])

        sigma_trend = pm.Exponential("sigma_trend", 1, dims=["shock_trend"])
        exog_ss_mod.build_statespace_graph(y)

    return m


@pytest.fixture(scope="session")
def idata(pymc_mod, rng, mock_pymc_sample):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        with pymc_mod:
            idata = pm.sample(
                draws=10,
                tune=1,
                chains=1,
                random_seed=rng,
                nuts_sampler="numpyro",
                progressbar=False,
            )
        with freeze_dims_and_data(pymc_mod):
            idata_prior = pm.sample_prior_predictive(
                samples=10, random_seed=rng, compile_kwargs={"mode": "JAX"}
            )

    idata.extend(idata_prior)
    return idata


@pytest.fixture(scope="session")
def idata_exog(exog_pymc_mod, rng, mock_pymc_sample):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        with exog_pymc_mod:
            idata = pm.sample(
                draws=10,
                tune=1,
                chains=1,
                random_seed=rng,
                nuts_sampler="numpyro",
                progressbar=False,
            )
        with freeze_dims_and_data(pymc_mod):
            idata_prior = pm.sample_prior_predictive(
                samples=10, random_seed=rng, compile_kwargs={"mode": "JAX"}
            )

    idata.extend(idata_prior)
    return idata


@pytest.mark.parametrize("group", ["posterior", "prior"])
@pytest.mark.parametrize("matrix", ALL_SAMPLE_OUTPUTS)
def test_no_nans_in_sampling_output(ss_mod, group, matrix, idata):
    assert not np.any(np.isnan(idata[group][matrix].values))


@pytest.mark.parametrize("group", ["prior", "posterior"])
@pytest.mark.parametrize("kind", ["conditional", "unconditional"])
def test_sampling_methods(group, kind, ss_mod, idata, rng):
    f = getattr(ss_mod, f"sample_{kind}_{group}")
    with pytest.warns(UserWarning, match="The RandomType SharedVariables"):
        test_idata = f(idata, random_seed=rng)

    if kind == "conditional":
        for output in ["filtered", "predicted", "smoothed"]:
            assert f"{output}_{group}" in test_idata
            assert not np.any(np.isnan(test_idata[f"{output}_{group}"].values))
            assert not np.any(np.isnan(test_idata[f"{output}_{group}_observed"].values))

    if kind == "unconditional":
        for output in ["latent", "observed"]:
            assert f"{group}_{output}" in test_idata
            assert not np.any(np.isnan(test_idata[f"{group}_{output}"].values))


@pytest.mark.parametrize("filter_output", ["predicted", "filtered", "smoothed"])
def test_forecast(filter_output, ss_mod, idata, rng):
    time_idx = idata.posterior.coords["time"].values

    with pytest.warns(UserWarning, match="The RandomType SharedVariables"):
        forecast_idata = ss_mod.forecast(
            idata, start=time_idx[-1], periods=10, filter_output=filter_output, random_seed=rng
        )

    assert forecast_idata.coords["time"].values.shape == (10,)
    assert forecast_idata.forecast_latent.dims == ("chain", "draw", "time", "state")
    assert forecast_idata.forecast_observed.dims == ("chain", "draw", "time", "observed_state")

    assert not np.any(np.isnan(forecast_idata.forecast_latent.values))
    assert not np.any(np.isnan(forecast_idata.forecast_observed.values))



import numpy as np
import pymc as pm
import pytest

from pymc_extras.model.transforms.autoreparam import vip_reparametrize


@pytest.fixture
def model_c():
    # TODO: Restructure tests so they check one dist at a time
    with pm.Model(coords=dict(a=range(5))) as mod:
        m = pm.Normal("m")
        s = pm.LogNormal("s")
        pm.Normal("g", m, s, dims="a")
        pm.Exponential("e", scale=s, shape=7)
    return mod


@pytest.fixture
def model_nc():
    with pm.Model(coords=dict(a=range(5))) as mod:
        m = pm.Normal("m")
        s = pm.LogNormal("s")
        pm.Deterministic("g", pm.Normal("z", dims="a") * s + m)
        pm.Deterministic("e", pm.Exponential("z_e", 1, shape=7) * s)
    return mod


@pytest.mark.parametrize("var", ["g", "e"])
def test_reparametrize_created(model_c: pm.Model, var):
    model_reparam, vip = vip_reparametrize(model_c, [var])
    assert f"{var}" in vip.get_lambda()
    assert f"{var}::lam_logit__" in model_reparam.named_vars
    assert f"{var}::tau_" in model_reparam.named_vars
    vip.set_all_lambda(1)
    assert ~np.isfinite(model_reparam[f"{var}::lam_logit__"].get_value()).any()


@pytest.mark.parametrize("var", ["g", "e"])
def test_random_draw(model_c: pm.Model, model_nc, var):
    model_c = pm.do(model_c, {"m": 3, "s": 2})
    model_nc = pm.do(model_nc, {"m": 3, "s": 2})
    model_v, vip = vip_reparametrize(model_c, [var])
    assert var in [v.name for v in model_v.deterministics]
    c = pm.draw(model_c[var], random_seed=42, draws=1000)
    nc = pm.draw(model_nc[var], random_seed=42, draws=1000)
    vip.set_all_lambda(1)
    v_1 = pm.draw(model_v[var], random_seed=42, draws=1000)
    vip.set_all_lambda(0)
    v_0 = pm.draw(model_v[var], random_seed=42, draws=1000)
    vip.set_all_lambda(0.5)
    v_05 = pm.draw(model_v[var], random_seed=42, draws=1000)
    np.testing.assert_allclose(c.mean(), nc.mean())
    np.testing.assert_allclose(c.mean(), v_0.mean())
    np.testing.assert_allclose(v_05.mean(), v_1.mean())
    np.testing.assert_allclose(v_1.mean(), nc.mean())

    np.testing.assert_allclose(c.std(), nc.std())
    np.testing.assert_allclose(c.std(), v_0.std())
    np.testing.assert_allclose(v_05.std(), v_1.std())
    np.testing.assert_allclose(v_1.std(), nc.std())


def test_reparam_fit(model_c):
    vars = ["g", "e"]
    model_v, vip = vip_reparametrize(model_c, ["g", "e"])
    with model_v:
        vip.fit(50000, random_seed=42)
    for var in vars:
        np.testing.assert_allclose(vip.get_lambda()[var], 0, atol=0.01)


def test_multilevel():
    with pm.Model(
        coords=dict(level=["Basement", "Floor"], county=[1, 2]),
    ) as model:
        # multilevel modelling
        a = pm.Normal("a")
        s = pm.HalfNormal("s")
        a_g = pm.Normal("a_g", a, s, shape=(2,), dims="level")
        s_g = pm.HalfNormal("s_g")
        a_ig = pm.Normal("a_ig", a_g, s_g, shape=(2, 2), dims=("county", "level"))

    model_r, vip = vip_reparametrize(model, ["a_g", "a_ig"])
    assert "a_g" in vip.get_lambda()
    assert "a_ig" in vip.get_lambda()
    assert {v.name for v in model_r.free_RVs} == {"a", "s", "a_g::tau_", "s_g", "a_ig::tau_"}
    assert "a_g" in [v.name for v in model_r.deterministics]


def test_set_truncate(model_c: pm.Model):
    model_v, vip = vip_reparametrize(model_c, ["m", "g"])
    vip.set_all_lambda(0.93)
    np.testing.assert_allclose(vip.get_lambda()["g"], 0.93)
    np.testing.assert_allclose(vip.get_lambda()["m"], 0.93)
    vip.truncate_all_lambda(0.1)
    np.testing.assert_allclose(vip.get_lambda()["g"], 1)
    np.testing.assert_allclose(vip.get_lambda()["m"], 1)

    vip.set_lambda(g=0.93, m=0.9)
    np.testing.assert_allclose(vip.get_lambda()["g"], 0.93)
    np.testing.assert_allclose(vip.get_lambda()["m"], 0.9)
    vip.truncate_lambda(g=0.2)
    np.testing.assert_allclose(vip.get_lambda()["g"], 1)
    np.testing.assert_allclose(vip.get_lambda()["m"], 0.9)


@pytest.mark.xfail(reason="FIX shape computation for lambda")
def test_lambda_shape():
    with pm.Model(coords=dict(a=[1, 2])) as model:
        b1 = pm.Normal("b1", dims="a")
        b2 = pm.Normal("b2", shape=2)
        b3 = pm.Normal("b3", size=2)
        b4 = pm.Normal("b4", np.asarray([1, 2]))
    model_v, vip = vip_reparametrize(model, ["b1", "b2", "b3", "b4"])
    lams = vip.get_lambda()
    for v in ["b1", "b2", "b3", "b4"]:
        assert lams[v].shape == (2,), v


@pytest.mark.xfail(reason="FIX shape computation for lambda")
def test_lambda_shape_transformed_1d():
    with pm.Model(coords=dict(a=[1, 2])) as model:
        b1 = pm.Exponential("b1", 1, dims="a")
        b2 = pm.Exponential("b2", 1, shape=2)
        b3 = pm.Exponential("b3", 1, size=2)
        b4 = pm.Exponential("b4", np.asarray([1, 2]))
    model_v, vip = vip_reparametrize(model, ["b1", "b2", "b3", "b4"])
    lams = vip.get_lambda()
    for v in ["b1", "b2", "b3", "b4"]:
        assert lams[v].shape == (2,), v



from itertools import combinations

import numpy as np
import pandas as pd
import pymc as pm
import pytensor
import pytensor.tensor as pt
import pytest
import statsmodels.api as sm

from numpy.testing import assert_allclose, assert_array_less
from pymc.testing import mock_sample_setup_and_teardown
from pytensor.graph.basic import explicit_graph_inputs

from pymc_extras.statespace import BayesianSARIMAX
from pymc_extras.statespace.models.utilities import (
    make_harvey_state_names,
    make_SARIMA_transition_matrix,
)
from pymc_extras.statespace.utils.constants import (
    SARIMAX_STATE_STRUCTURES,
    SHORT_NAME_TO_LONG,
)
from tests.statespace.shared_fixtures import (  # pylint: disable=unused-import
    rng,
)
from tests.statespace.test_utilities import (
    load_nile_test_data,
    make_stationary_params,
    simulate_from_numpy_model,
)

mock_sample = pytest.fixture()(mock_sample_setup_and_teardown)

floatX = pytensor.config.floatX
ATOL = 1e-8 if floatX.endswith("64") else 1e-6
RTOL = 0 if floatX.endswith("64") else 1e-6

test_state_names = [
    ["data", "state_1", "state_2"],
    ["data", "data_star", "state_star_1", "state_star_2"],
    ["data", "D1.data", "data_star", "state_star_1", "state_star_2"],
    ["data", "D1.data", "D1^2.data", "data_star", "state_star_1", "state_star_2"],
    ["data", "state_1", "state_2", "state_3"],
    ["data", "state_1", "state_2", "state_3", "state_4", "state_5", "state_6", "state_7"],
    [
        "data",
        "state_1",
        "state_2",
        "state_3",
        "state_4",
        "state_5",
        "state_6",
        "state_7",
        "state_8",
        "state_9",
        "state_10",
    ],
    [
        "data",
        "L1.data",
        "L2.data",
        "L3.data",
        "data_star",
        "state_star_1",
        "state_star_2",
        "state_star_3",
    ],
    [
        "data",
        "L1.data",
        "L2.data",
        "L3.data",
        "D4.data",
        "L1D4.data",
        "L2D4.data",
        "L3D4.data",
        "data_star",
        "state_star_1",
        "state_star_2",
        "state_star_3",
    ],
    [
        "data",
        "D1.data",
        "L1D1.data",
        "L2D1.data",
        "L3D1.data",
        "data_star",
        "state_star_1",
        "state_star_2",
        "state_star_3",
        "state_star_4",
        "state_star_5",
    ],
    [
        "data",
        "D1.data",
        "D1^2.data",
        "L1D1^2.data",
        "L2D1^2.data",
        "L3D1^2.data",
        "data_star",
        "state_star_1",
        "state_star_2",
        "state_star_3",
        "state_star_4",
        "state_star_5",
    ],
    [
        "data",
        "D1.data",
        "D1^2.data",
        "L1D1^2.data",
        "L2D1^2.data",
        "L3D1^2.data",
        "D1^2D4.data",
        "L1D1^2D4.data",
        "L2D1^2D4.data",
        "L3D1^2D4.data",
        "data_star",
        "state_star_1",
        "state_star_2",
        "state_star_3",
        "state_star_4",
        "state_star_5",
    ],
    [
        "data",
        "D1.data",
        "L1D1.data",
        "L2D1.data",
        "D1D3.data",
        "L1D1D3.data",
        "L2D1D3.data",
        "D1D3^2.data",
        "L1D1D3^2.data",
        "L2D1D3^2.data",
        "data_star",
        "state_star_1",
        "state_star_2",
        "state_star_3",
        "state_star_4",
    ],
    ["data", "data_star"] + [f"state_star_{i + 1}" for i in range(26)],
]

test_orders = [
    (2, 0, 2, 0, 0, 0, 0),
    (2, 1, 2, 0, 0, 0, 0),
    (2, 2, 2, 0, 0, 0, 0),
    (2, 3, 2, 0, 0, 0, 0),
    (0, 0, 0, 1, 0, 0, 4),
    (0, 0, 0, 2, 0, 1, 4),
    (2, 0, 2, 2, 0, 2, 4),
    (0, 0, 0, 1, 1, 0, 4),
    (0, 0, 0, 1, 2, 0, 4),
    (1, 1, 1, 1, 1, 1, 4),
    (1, 2, 1, 1, 1, 1, 4),
    (1, 2, 1, 1, 2, 1, 4),
    (1, 1, 1, 1, 3, 1, 3),
    (2, 1, 2, 2, 0, 2, 12),
]

ids = [f"p={p},d={d},q={q},P={P},D={D},Q={Q},S={S}" for (p, d, q, P, D, Q, S) in test_orders]


@pytest.fixture
def data():
    return load_nile_test_data()


@pytest.fixture(scope="session")
def arima_mod():
    return BayesianSARIMAX(order=(2, 0, 1), stationary_initialization=True, verbose=False)


@pytest.fixture(scope="session")
def pymc_mod(arima_mod):
    data = load_nile_test_data()

    with pm.Model(coords=arima_mod.coords) as pymc_mod:
        # x0  = pm.Normal('x0', dims=['state'])
        # P0_diag = pm.Gamma('P0_diag', alpha=2, beta=1, dims=['state'])
        # P0 = pm.Deterministic('P0', pt.diag(P0_diag), dims=['state', 'state_aux'])
        ar_params = pm.Normal("ar_params", sigma=0.1, dims=["lag_ar"])
        ma_params = pm.Normal("ma_params", sigma=1, dims=["lag_ma"])
        sigma_state = pm.Exponential("sigma_state", 0.5)
        arima_mod.build_statespace_graph(data=data, save_kalman_filter_outputs_in_idata=True)

    return pymc_mod


@pytest.fixture(scope="session")
def arima_mod_interp():
    return BayesianSARIMAX(
        order=(3, 0, 3),
        stationary_initialization=False,
        verbose=False,
        state_structure="interpretable",
        measurement_error=True,
    )


@pytest.fixture(scope="session")
def pymc_mod_interp(arima_mod_interp):
    data = load_nile_test_data()

    with pm.Model(coords=arima_mod_interp.coords) as pymc_mod:
        x0 = pm.Normal("x0", dims=["state"])
        P0_sigma = pm.Exponential("P0_sigma", 1)
        P0 = pm.Deterministic(
            "P0", pt.eye(arima_mod_interp.k_states) * P0_sigma, dims=["state", "state_aux"]
        )
        ar_params = pm.Normal("ar_params", sigma=0.1, dims=["lag_ar"])
        ma_params = pm.Normal("ma_params", sigma=1, dims=["lag_ma"])
        sigma_state = pm.Exponential("sigma_state", 0.5)
        sigma_obs = pm.Exponential("sigma_obs", 0.1)

        arima_mod_interp.build_statespace_graph(data=data, save_kalman_filter_outputs_in_idata=True)

    return pymc_mod


def test_mode_argument():
    # Mode argument should be passed to the parent class
    mod = BayesianSARIMAX(order=(0, 0, 3), mode="FAST_RUN", verbose=False)
    assert mod.mode == "FAST_RUN"


@pytest.mark.parametrize(
    "p,d,q,P,D,Q,S,expected_names",
    [(*order, name) for order, name in zip(test_orders, test_state_names)],
    ids=ids,
)
def test_harvey_state_names(p, d, q, P, D, Q, S, expected_names):
    if all([x == 0 for x in [p, d, q, P, D, Q, S]]):
        pytest.skip("Skip all zero case")

    k_states = max(p + P * S, q + Q * S + 1) + (S * D + d)
    states = make_harvey_state_names(p, d, q, P, D, Q, S)

    assert len(states) == k_states
    missing_from_expected = set(expected_names) - set(states)
    assert len(missing_from_expected) == 0


@pytest.mark.parametrize("p,d,q,P,D,Q,S", test_orders)
def test_make_SARIMA_transition_matrix(p, d, q, P, D, Q, S):
    T = make_SARIMA_transition_matrix(p, d, q, P, D, Q, S)
    mod = sm.tsa.SARIMAX(np.random.normal(size=100), order=(p, d, q), seasonal_order=(P, D, Q, S))
    T2 = mod.ssm["transition"]

    if D > 2:
        pytest.skip("Statsmodels has a bug when D > 2, skip this test.")
    else:
        assert_allclose(T, T2, err_msg="Transition matrix does not match statsmodels")


@pytest.mark.parametrize("p, d, q, P, D, Q, S", test_orders, ids=ids)
@pytest.mark.filterwarnings(
    "ignore:Non-invertible starting MA parameters found.",
    "ignore:Non-stationary starting autoregressive parameters found",
    "ignore:Non-invertible starting seasonal moving average",
    "ignore:Non-stationary starting seasonal autoregressive",
    "ignore:divide by zero encountered in matmul:RuntimeWarning",
    "ignore:overflow encountered in matmul:RuntimeWarning",
    "ignore:invalid value encountered in matmul:RuntimeWarning",
)
def test_SARIMAX_update_matches_statsmodels(p, d, q, P, D, Q, S, data, rng):
    sm_sarimax = sm.tsa.SARIMAX(data, order=(p, d, q), seasonal_order=(P, D, Q, S))

    param_names = sm_sarimax.param_names
    param_d = {name: getattr(np, floatX)(rng.normal(scale=0.1) ** 2) for name in param_names}

    res = sm_sarimax.fit_constrained(param_d)
    mod = BayesianSARIMAX(
        order=(p, d, q), seasonal_order=(P, D, Q, S), verbose=False, stationary_initialization=False
    )

    with pm.Model() as pm_mod:
        x0 = pm.Normal("x0", shape=(mod.k_states,))
        P0 = pm.Deterministic("P0", pt.eye(mod.k_states, dtype=floatX))

        if q > 0:
            pm.Deterministic(
                "ma_params",
                pt.as_tensor_variable(
                    np.array([param_d[k] for k in param_d if k.startswith("ma.") and "S." not in k])
                ),
            )
        if p > 0:
            pm.Deterministic(
                "ar_params",
                pt.as_tensor_variable(
                    np.array([param_d[k] for k in param_d if k.startswith("ar.") and "S." not in k])
                ),
            )
        if P > 0:
            pm.Deterministic(
                "seasonal_ar_params",
                pt.as_tensor_variable(
                    np.array([param_d[k] for k in param_d if k.startswith("ar.S.")])
                ),
            )

        if Q > 0:
            pm.Deterministic(
                "seasonal_ma_params",
                pt.as_tensor_variable(
                    np.array([param_d[k] for k in param_d if k.startswith("ma.S.")])
                ),
            )

        pm.Deterministic("sigma_state", pt.as_tensor_variable(np.sqrt(param_d["sigma2"])))

        mod._insert_random_variables()
        matrices = pm.draw(mod.subbed_ssm)
        matrix_dict = dict(zip(SHORT_NAME_TO_LONG.values(), matrices))

    for matrix in ["transition", "selection", "state_cov", "obs_cov", "design"]:
        if matrix == "transition" and D > 2:
            pytest.skip("Statsmodels has a bug when D > 2, skip this test.)")
        assert_allclose(matrix_dict[matrix], sm_sarimax.ssm[matrix], err_msg=f"{matrix} not equal")


@pytest.mark.parametrize("filter_output", ["filtered", "predicted", "smoothed"])
def test_all_prior_covariances_are_PSD(filter_output, pymc_mod, rng):
    rv = pymc_mod[f"{filter_output}_covariances"]
    cov_mats = pm.draw(rv, 100, random_seed=rng)
    w, v = np.linalg.eig(cov_mats)
    assert_array_less(0, w, err_msg=f"Smallest eigenvalue: {min(w.ravel())}")


def test_interpretable_raises_if_d_nonzero():
    with pytest.raises(
        ValueError, match="Cannot use interpretable state structure with statespace differencing"
    ):
        BayesianSARIMAX(
            order=(2, 1, 1),
            stationary_initialization=False,
            verbose=False,
            state_structure="interpretable",
        )


def test_interpretable_states_are_interpretable(arima_mod_interp, pymc_mod_interp):
    with pymc_mod_interp:
        prior = pm.sample_prior_predictive(draws=10)

    prior_outputs = arima_mod_interp.sample_unconditional_prior(prior)
    ar_lags = prior.prior.coords["lag_ar"].values - 1
    ma_lags = prior.prior.coords["lag_ma"].values - 1

    # Check the first p states are lags of the previous state
    for t, tm1 in zip(ar_lags[1:], ar_lags[:-1]):
        assert_allclose(
            prior_outputs.prior_latent.isel(state=t).values[1:],
            prior_outputs.prior_latent.isel(state=tm1).values[:-1],
            err_msg=f"State {tm1} is not a lagged version of state {t} (AR lags)",
        )

    # Check the next p+q states are lags of the innovations
    n = len(ar_lags)
    for t, tm1 in zip(ma_lags[1:], ma_lags[:-1]):
        assert_allclose(
            prior_outputs.prior_latent.isel(state=n + t).values[1:],
            prior_outputs.prior_latent.isel(state=n + tm1).values[:-1],
            err_msg=f"State {n + tm1} is not a lagged version of state {n + t} (MA lags)",
        )


@pytest.mark.parametrize("p, d, q, P, D, Q, S", test_orders, ids=ids)
@pytest.mark.filterwarnings(
    "ignore:Non-invertible starting MA parameters found.",
    "ignore:Non-stationary starting autoregressive parameters found",
    "ignore:Maximum Likelihood optimization failed to converge.",
    "ignore:divide by zero encountered in matmul:RuntimeWarning",
    "ignore:overflow encountered in matmul:RuntimeWarning",
    "ignore:invalid value encountered in matmul:RuntimeWarning",
)
def test_representations_are_equivalent(p, d, q, P, D, Q, S, data, rng):
    if (d + D) > 0:
        pytest.skip('state_structure = "interpretable" cannot include statespace differences')

    shared_params = make_stationary_params(data, p, d, q, P, D, Q, S)
    test_values = {}

    for representation in SARIMAX_STATE_STRUCTURES:
        rng = np.random.default_rng(sum(map(ord, "representation test")))
        mod = BayesianSARIMAX(
            order=(p, d, q),
            seasonal_order=(P, D, Q, S),
            stationary_initialization=False,
            verbose=False,
            state_structure=representation,
        )
        shared_params.update(
            {
                "x0": np.zeros(mod.k_states, dtype=floatX),
                "initial_state_cov": np.eye(mod.k_states, dtype=floatX) * 100,
            }
        )
        x, y = simulate_from_numpy_model(mod, rng, shared_params)
        test_values[representation] = y

    all_pairs = combinations(SARIMAX_STATE_STRUCTURES, r=2)
    for rep_1, rep_2 in all_pairs:
        assert_allclose(
            test_values[rep_1],
            test_values[rep_2],
            err_msg=f"{rep_1} and {rep_2} are not the same",
            atol=ATOL,
            rtol=RTOL,
        )


@pytest.mark.parametrize("order, name", [((4, 1, 0, 0), "AR"), ((0, 0, 4, 1), "MA")])
def test_invalid_order_raises(order, name):
    p, P, q, Q = order
    with pytest.raises(ValueError, match=f"The following {name} and seasonal {name} terms overlap"):
        BayesianSARIMAX(order=(p, 0, q), seasonal_order=(P, 0, Q, 4))


def test_SARIMA_with_exogenous(rng, mock_sample):
    ss_mod = BayesianSARIMAX(order=(3, 0, 1), seasonal_order=(1, 0, 0, 12), k_exog=2)

    assert ss_mod.param_dims["beta_exog"] == ("exogenous",)
    assert ss_mod.data_names == ["exogenous_data"]
    assert ss_mod.coords["exogenous"] == ["exogenous_0", "exogenous_1"]

    obs_intercept = ss_mod.ssm["obs_intercept"]
    assert obs_intercept.type.shape == (None, ss_mod.k_endog)

    intercept_fn = pytensor.function(
        inputs=list(explicit_graph_inputs(obs_intercept)), outputs=obs_intercept
    )
    data_val = rng.normal(size=(100, 2)).astype(floatX)
    beta_val = rng.normal(size=(2,)).astype(floatX)

    intercept_val = intercept_fn(data_val, beta_val)
    np.testing.assert_allclose(intercept_val, intercept_fn(data_val, beta_val))

    data_df = pd.DataFrame(
        rng.normal(size=(100, 1)),
        index=pd.date_range(start="2020-01-01", periods=100, freq="D"),
        columns=["endog"],
    )

    with pm.Model(coords=ss_mod.coords) as pymc_mod:
        pm.Data("exogenous_data", data_val, dims=["time", "exogenous"])

        ar_params = pm.Normal("ar_params", dims=["lag_ar"])
        ma_params = pm.Normal("ma_params", dims=["lag_ma"])
        seasonal_ar_params = pm.Normal("seasonal_ar_params", dims=["seasonal_lag_ar"])

        beta_exog = pm.Normal("beta_exog", dims=["exogenous"])

        sigma_state = pm.Exponential("sigma_state", 1.0)

        ss_mod.build_statespace_graph(data=data_df, save_kalman_filter_outputs_in_idata=True)
        idata = pm.sample(chains=2, draws=100)

    assert "exogenous_data" in idata.constant_data
    assert idata.posterior.beta_exog.shape == (
        2,
        100,
        2,
    )
    np.testing.assert_allclose(ss_mod._fit_exog_data["exogenous_data"]["value"], data_val)



from functools import partial

import numpy as np
import pandas as pd
import pymc as pm
import pytest

from numpy.testing import assert_allclose
from pytensor import config
from pytensor import tensor as pt
from scipy import linalg

from pymc_extras.statespace.models import structural as st
from tests.statespace.test_utilities import unpack_symbolic_matrices_with_params

floatX = config.floatX
ATOL = 1e-8 if floatX.endswith("64") else 1e-4
RTOL = 0 if floatX.endswith("64") else 1e-6


def test_add_components():
    ll = st.LevelTrendComponent(order=2)
    se = st.TimeSeasonality(name="seasonal", season_length=12)
    mod = ll + se

    ll_params = {
        "initial_level_trend": np.zeros(2, dtype=floatX),
        "sigma_level_trend": np.ones(2, dtype=floatX),
    }
    se_params = {
        "params_seasonal": np.ones(11, dtype=floatX),
        "sigma_seasonal": 1.0,
    }
    all_params = ll_params.copy()
    all_params.update(se_params)

    (ll_x0, ll_P0, ll_c, ll_d, ll_T, ll_Z, ll_R, ll_H, ll_Q) = unpack_symbolic_matrices_with_params(
        ll, ll_params
    )
    (se_x0, se_P0, se_c, se_d, se_T, se_Z, se_R, se_H, se_Q) = unpack_symbolic_matrices_with_params(
        se, se_params
    )
    x0, P0, c, d, T, Z, R, H, Q = unpack_symbolic_matrices_with_params(mod, all_params)

    for property in ["param_names", "shock_names", "param_info", "coords", "param_dims"]:
        assert [x in getattr(mod, property) for x in getattr(ll, property)]
        assert [x in getattr(mod, property) for x in getattr(se, property)]

    assert (mod.observed_state_names == ll.observed_state_names) and (
        ll.observed_state_names == se.observed_state_names
    )

    ll_mats = [ll_T, ll_R, ll_Q]
    se_mats = [se_T, se_R, se_Q]
    all_mats = [T, R, Q]

    for ll_mat, se_mat, all_mat in zip(ll_mats, se_mats, all_mats):
        assert_allclose(all_mat, linalg.block_diag(ll_mat, se_mat), atol=ATOL, rtol=RTOL)

    ll_mats = [ll_x0, ll_c, ll_Z]
    se_mats = [se_x0, se_c, se_Z]
    all_mats = [x0, c, Z]
    axes = [0, 0, 1]

    for ll_mat, se_mat, all_mat, axis in zip(ll_mats, se_mats, all_mats, axes):
        assert_allclose(all_mat, np.concatenate([ll_mat, se_mat], axis=axis), atol=ATOL, rtol=RTOL)


def test_add_components_multiple_observed():
    ll = st.LevelTrendComponent(order=2, observed_state_names=["data_1", "data_2"])
    me = st.MeasurementError(name="obs", observed_state_names=["data_1", "data_2"])

    mod = (ll + me).build()

    for property in ["param_names", "shock_names", "param_info", "coords", "param_dims"]:
        assert [x in getattr(mod, property) for x in getattr(ll, property)]


@pytest.mark.skipif(floatX.endswith("32"), reason="Prior covariance not PSD at half-precision")
def test_extract_components_from_idata(rng):
    time_idx = pd.date_range(start="2000-01-01", freq="D", periods=100)
    data = pd.DataFrame(rng.normal(size=(100, 2)), columns=["a", "b"], index=time_idx)

    y = pd.DataFrame(rng.normal(size=(100, 1)), columns=["data"], index=time_idx)

    ll = st.LevelTrendComponent()
    season = st.FrequencySeasonality(name="seasonal", season_length=12, n=2, innovations=False)
    reg = st.RegressionComponent(state_names=["a", "b"], name="exog")
    me = st.MeasurementError("obs")
    mod = (ll + season + reg + me).build(verbose=False)

    with pm.Model(coords=mod.coords) as m:
        data_exog = pm.Data("data_exog", data.values)

        x0 = pm.Normal("x0", dims=["state"])
        P0 = pm.Deterministic("P0", pt.eye(mod.k_states), dims=["state", "state_aux"])
        beta_exog = pm.Normal("beta_exog", dims=["state_exog"])
        initial_trend = pm.Normal("initial_level_trend", dims=["state_level_trend"])
        sigma_trend = pm.Exponential("sigma_level_trend", 1, dims=["shock_level_trend"])
        seasonal_coefs = pm.Normal("params_seasonal", dims=["state_seasonal"])
        sigma_obs = pm.Exponential("sigma_obs", 1)

        mod.build_statespace_graph(y)

        prior = pm.sample_prior_predictive(draws=10)

    filter_prior = mod.sample_conditional_prior(prior)
    comp_prior = mod.extract_components_from_idata(filter_prior)
    comp_states = comp_prior.filtered_prior.coords["state"].values
    expected_states = ["level_trend[level]", "level_trend[trend]", "seasonal", "exog[a]", "exog[b]"]
    missing = set(comp_states) - set(expected_states)

    assert len(missing) == 0, missing


def test_extract_multiple_observed(rng):
    time_idx = pd.date_range(start="2000-01-01", freq="D", periods=100)
    data = pd.DataFrame(rng.normal(size=(100, 2)), columns=["a", "b"], index=time_idx)

    y = pd.DataFrame(
        rng.normal(size=(100, 3)), columns=["data_1", "data_2", "data_3"], index=time_idx
    )

    ll = st.LevelTrendComponent(name="trend", observed_state_names=["data_1", "data_2"])
    season = st.FrequencySeasonality(
        name="seasonal", observed_state_names=["data_1"], season_length=12, n=2, innovations=False
    )
    reg = st.RegressionComponent(
        state_names=["a", "b"], name="exog", observed_state_names=["data_2", "data_3"]
    )
    ar = st.AutoregressiveComponent(observed_state_names=["data_1", "data_2"], order=3)
    me = st.MeasurementError("obs", observed_state_names=["data_1", "data_3"])
    mod = (ll + season + reg + ar + me).build(verbose=True)

    with pm.Model(coords=mod.coords) as m:
        data_exog = pm.Data("data_exog", data.values)

        x0 = pm.Normal("x0", dims=["state"])
        P0 = pm.Deterministic("P0", pt.eye(mod.k_states), dims=["state", "state_aux"])
        beta_exog = pm.Normal("beta_exog", dims=["endog_exog", "state_exog"])
        params_auto_regressive = pm.Normal(
            "params_auto_regressive", dims=["endog_auto_regressive", "lag_auto_regressive"]
        )
        sigma_auto_regressive = pm.Normal("sigma_auto_regressive", dims=["endog_auto_regressive"])
        initial_trend = pm.Normal("initial_trend", dims=["endog_trend", "state_trend"])
        sigma_trend = pm.Exponential("sigma_trend", 1, dims=["endog_trend", "shock_trend"])
        seasonal_coefs = pm.Normal("params_seasonal", dims=["state_seasonal"])
        sigma_obs = pm.Exponential("sigma_obs", 1, dims=["endog_obs"])

        mod.build_statespace_graph(y)

        prior = pm.sample_prior_predictive(draws=10)

    filter_prior = mod.sample_conditional_prior(prior)
    comp_prior = mod.extract_components_from_idata(filter_prior)
    comp_states = comp_prior.filtered_prior.coords["state"].values

    expected_states = [
        "trend[level[data_1]]",
        "trend[trend[data_1]]",
        "trend[level[data_2]]",
        "trend[trend[data_2]]",
        "seasonal[data_1]",
        "exog[a[data_2]]",
        "exog[b[data_2]]",
        "exog[a[data_3]]",
        "exog[b[data_3]]",
        "auto_regressive[data_1]",
        "auto_regressive[data_2]",
    ]

    missing = set(comp_states) - set(expected_states)
    assert len(missing) == 0, missing


@pytest.mark.parametrize(
    "arg_type", [tuple, list, set, np.array], ids=["tuple", "list", "set", "array"]
)
def test_sequence_type_component_arguments(arg_type):
    state_names = list("ABCDEFG")
    components = [
        st.LevelTrendComponent,
        partial(st.CycleComponent, cycle_length=12),
        st.AutoregressiveComponent,
        partial(st.FrequencySeasonality, season_length=12),
        partial(st.TimeSeasonality, season_length=12),
        st.MeasurementError,
    ]

    components = [
        components[i](observed_state_names=arg_type(state_names))
        for i in np.random.choice(len(components), size=3, replace=False)
    ]
    ss_mod = sum(components[1:], start=components[0]).build(verbose=False)

    assert ss_mod.k_endog == len(state_names)
    assert sorted(ss_mod.observed_states) == sorted(list(state_names))



import itertools

from collections.abc import Sequence
from itertools import zip_longest

from pymc import SymbolicRandomVariable
from pymc.model.fgraph import ModelVar
from pymc.variational.minibatch_rv import MinibatchRandomVariable
from pytensor.graph import Variable, ancestors
from pytensor.graph.basic import io_toposort
from pytensor.tensor import TensorType, TensorVariable
from pytensor.tensor.blockwise import Blockwise
from pytensor.tensor.elemwise import CAReduce, DimShuffle, Elemwise
from pytensor.tensor.random.op import RandomVariable
from pytensor.tensor.rewriting.subtensor import is_full_slice
from pytensor.tensor.shape import Shape
from pytensor.tensor.subtensor import AdvancedSubtensor, Subtensor, get_idx_list
from pytensor.tensor.type_other import NoneTypeT

from pymc_extras.model.marginal.distributions import MarginalRV


def static_shape_ancestors(vars):
    """Identify ancestors Shape Ops of static shapes (therefore constant in a valid graph)."""
    return [
        var
        for var in ancestors(vars)
        if (
            var.owner
            and isinstance(var.owner.op, Shape)
            # All static dims lengths of Shape input are known
            and None not in var.owner.inputs[0].type.shape
        )
    ]


def find_conditional_input_rvs(output_rvs, all_rvs):
    """Find conditionally indepedent input RVs."""
    other_rvs = [other_rv for other_rv in all_rvs if other_rv not in output_rvs]
    blockers = other_rvs + static_shape_ancestors(tuple(all_rvs) + tuple(output_rvs))
    return [var for var in ancestors(output_rvs, blockers=blockers) if var in other_rvs]


def is_conditional_dependent(
    dependent_rv: TensorVariable, dependable_rv: TensorVariable, all_rvs
) -> bool:
    """Check if dependent_rv is conditionall dependent on dependable_rv,
    given all conditionally independent all_rvs"""

    return dependable_rv in find_conditional_input_rvs((dependent_rv,), all_rvs)


def find_conditional_dependent_rvs(dependable_rv, all_rvs):
    """Find rvs than depend on dependable"""
    return [
        rv
        for rv in all_rvs
        if (rv is not dependable_rv and is_conditional_dependent(rv, dependable_rv, all_rvs))
    ]


def get_support_axes(op) -> tuple[tuple[int, ...], ...]:
    if isinstance(op, MarginalRV):
        return op.support_axes
    else:
        # For vanilla RVs, the support axes are the last ndim_supp
        return (tuple(range(-op.ndim_supp, 0)),)


def _advanced_indexing_axis_and_ndim(idxs) -> tuple[int, int]:
    """Find the output axis and dimensionality of the advanced indexing group (i.e., array indexing).

    There is a special case: when there are non-consecutive advanced indexing groups, the advanced indexing
    group is always moved to the front.

    See: https://numpy.org/doc/stable/user/basics.indexing.html#combining-advanced-and-basic-indexing
    """
    adv_group_axis = None
    simple_group_after_adv = False
    for axis, idx in enumerate(idxs):
        if isinstance(idx.type, TensorType):
            if simple_group_after_adv:
                # Special non-consecutive case
                adv_group_axis = 0
                break
            elif adv_group_axis is None:
                adv_group_axis = axis
        elif adv_group_axis is not None:
            # Special non-consecutive case
            simple_group_after_adv = True

    adv_group_ndim = max(idx.type.ndim for idx in idxs if isinstance(idx.type, TensorType))
    return adv_group_axis, adv_group_ndim


DIMS = tuple[int | None, ...]
VAR_DIMS = dict[Variable, DIMS]


def _broadcast_dims(
    inputs_dims: Sequence[DIMS],
) -> DIMS:
    output_ndim = max((len(input_dim) for input_dim in inputs_dims), default=0)

    # Add missing dims
    inputs_dims = [
        (None,) * (output_ndim - len(input_dim)) + input_dim for input_dim in inputs_dims
    ]

    # Find which known dims show in the output, while checking no mixing
    output_dims = []
    for inputs_dim in zip(*inputs_dims):
        output_dim = None
        for input_dim in inputs_dim:
            if input_dim is None:
                continue
            if output_dim is not None and output_dim != input_dim:
                raise ValueError("Different known dimensions mixed via broadcasting")
            output_dim = input_dim
        output_dims.append(output_dim)

    # Check for duplicates
    known_dims = [dim for dim in output_dims if dim is not None]
    if len(known_dims) > len(set(known_dims)):
        raise ValueError("Same known dimension used in different axis after broadcasting")

    return tuple(output_dims)


def _subgraph_batch_dim_connection(var_dims: VAR_DIMS, input_vars, output_vars) -> VAR_DIMS:
    for node in io_toposort(input_vars, output_vars):
        inputs_dims = [
            var_dims.get(inp, ((None,) * inp.type.ndim) if hasattr(inp.type, "ndim") else ())
            for inp in node.inputs
        ]

        if all(dim is None for input_dims in inputs_dims for dim in input_dims):
            # None of the inputs are related to the batch_axes of the input_vars
            continue

        elif isinstance(node.op, ModelVar):
            var_dims[node.outputs[0]] = inputs_dims[0]

        elif isinstance(node.op, DimShuffle):
            [input_dims] = inputs_dims
            output_dims = tuple(None if i == "x" else input_dims[i] for i in node.op.new_order)
            var_dims[node.outputs[0]] = output_dims

        elif isinstance(node.op, MarginalRV) or (
            isinstance(node.op, SymbolicRandomVariable) and node.op.extended_signature is None
        ):
            # MarginalRV and SymbolicRandomVariables without signature are a wild-card,
            # so we need to introspect the inner graph.
            op = node.op
            inner_inputs = op.inner_inputs
            inner_outputs = op.inner_outputs

            inner_var_dims = _subgraph_batch_dim_connection(
                dict(zip(inner_inputs, inputs_dims)), inner_inputs, inner_outputs
            )

            support_axes = iter(get_support_axes(op))
            if isinstance(op, MarginalRV):
                # The first output is the marginalized variable for which we don't compute support axes
                support_axes = itertools.chain(((),), support_axes)
            for i, (out, inner_out) in enumerate(zip(node.outputs, inner_outputs)):
                if not isinstance(out.type, TensorType):
                    continue
                support_axes_out = next(support_axes)

                if inner_out in inner_var_dims:
                    out_dims = inner_var_dims[inner_out]
                    if any(
                        dim is not None for dim in (out_dims[axis] for axis in support_axes_out)
                    ):
                        raise ValueError(f"Known dim corresponds to core dimension of {node.op}")
                    var_dims[out] = out_dims

        elif isinstance(node.op, Elemwise | Blockwise | RandomVariable | SymbolicRandomVariable):
            # NOTE: User-provided CustomDist may not respect core dimensions on the left.

            if isinstance(node.op, Elemwise):
                op_batch_ndim = node.outputs[0].type.ndim
            else:
                op_batch_ndim = node.op.batch_ndim(node)

            if isinstance(node.op, SymbolicRandomVariable):
                # SymbolicRandomVariable don't have explicit expand_dims unlike the other Ops considered in this
                [_, _, param_idxs], _ = node.op.get_input_output_type_idxs(
                    node.op.extended_signature
                )
                for param_idx, param_core_ndim in zip(param_idxs, node.op.ndims_params):
                    param_dims = inputs_dims[param_idx]
                    missing_ndim = op_batch_ndim - (len(param_dims) - param_core_ndim)
                    inputs_dims[param_idx] = (None,) * missing_ndim + param_dims

            if any(
                dim is not None for input_dim in inputs_dims for dim in input_dim[op_batch_ndim:]
            ):
                raise ValueError(
                    f"Use of known dimensions as core dimensions of op {node.op} not supported."
                )

            batch_dims = _broadcast_dims(
                tuple(input_dims[:op_batch_ndim] for input_dims in inputs_dims)
            )
            for out in node.outputs:
                if isinstance(out.type, TensorType):
                    core_ndim = out.type.ndim - op_batch_ndim
                    output_dims = batch_dims + (None,) * core_ndim
                    var_dims[out] = output_dims

        elif isinstance(node.op, CAReduce):
            [input_dims] = inputs_dims

            axes = node.op.axis
            if isinstance(axes, int):
                axes = (axes,)
            elif axes is None:
                axes = tuple(range(node.inputs[0].type.ndim))

            if any(input_dims[axis] for axis in axes):
                raise ValueError(
                    f"Use of known dimensions as reduced dimensions of op {node.op} not supported."
                )

            output_dims = [dims for i, dims in enumerate(input_dims) if i not in axes]
            var_dims[node.outputs[0]] = tuple(output_dims)

        elif isinstance(node.op, Subtensor):
            value_dims, *keys_dims = inputs_dims
            # Dims in basic indexing must belong to the value variable, since indexing keys are always scalar
            assert not any(dim is None for dim in keys_dims)
            keys = get_idx_list(node.inputs, node.op.idx_list)

            output_dims = []
            for value_dim, idx in zip_longest(value_dims, keys, fillvalue=slice(None)):
                if idx == slice(None):
                    # Dim is kept
                    output_dims.append(value_dim)
                elif value_dim is not None:
                    raise ValueError(
                        "Partial slicing or indexing of known dimensions not supported."
                    )
                elif isinstance(idx, slice):
                    # Unknown dimensions kept by partial slice.
                    output_dims.append(None)

            var_dims[node.outputs[0]] = tuple(output_dims)

        elif isinstance(node.op, AdvancedSubtensor):
            # AdvancedSubtensor dimensions can show up as both the indexed variable and indexing variables
            value, *keys = node.inputs
            value_dims, *keys_dims = inputs_dims

            # Just to stay sane, we forbid any boolean indexing...
            if any(isinstance(idx.type, TensorType) and idx.type.dtype == "bool" for idx in keys):
                raise NotImplementedError(
                    f"Array indexing with boolean variables in node {node} not supported."
                )

            if any(dim is not None for dim in value_dims) and any(
                dim is not None for key_dims in keys_dims for dim in key_dims
            ):
                # Both indexed variable and indexing variables have known dimensions
                # I am to lazy to think through these, so we raise for now.
                raise NotImplementedError(
                    f"Simultaneous use of known dimensions in indexed and indexing variables in node {node} not supported."
                )

            adv_group_axis, adv_group_ndim = _advanced_indexing_axis_and_ndim(keys)

            if any(dim is not None for dim in value_dims):
                # Indexed variable has known dimensions

                if any(isinstance(idx.type, NoneTypeT) for idx in keys):
                    # Corresponds to an expand_dims, for now not supported
                    raise NotImplementedError(
                        f"Advanced indexing in node {node} which introduces new axis is not supported."
                    )

                non_adv_dims = []
                for value_dim, idx in zip_longest(value_dims, keys, fillvalue=slice(None)):
                    if is_full_slice(idx):
                        non_adv_dims.append(value_dim)
                    elif value_dim is not None:
                        # We are trying to partially slice or index a known dimension
                        raise ValueError(
                            "Partial slicing or advanced integer indexing of known dimensions not supported."
                        )
                    elif isinstance(idx, slice):
                        # Unknown dimensions kept by partial slice.
                        non_adv_dims.append(None)

                # Insert unknown dimensions corresponding to advanced indexing
                output_dims = tuple(
                    non_adv_dims[:adv_group_axis]
                    + [None] * adv_group_ndim
                    + non_adv_dims[adv_group_axis:]
                )

            else:
                # Indexing keys have known dimensions.
                # Only array indices can have dimensions, the rest are just slices or newaxis

                # Advanced indexing variables broadcast together, so we apply same rules as in Elemwise
                adv_dims = _broadcast_dims(keys_dims)

                start_non_adv_dims = (None,) * adv_group_axis
                end_non_adv_dims = (None,) * (
                    node.outputs[0].type.ndim - adv_group_axis - adv_group_ndim
                )
                output_dims = start_non_adv_dims + adv_dims + end_non_adv_dims

            var_dims[node.outputs[0]] = output_dims

        elif isinstance(node.op, MinibatchRandomVariable):
            var_dims[node.outputs[0]] = inputs_dims[0]

        else:
            raise NotImplementedError(f"Marginalization through operation {node} not supported.")

    return var_dims


def subgraph_batch_dim_connection(input_var, output_vars) -> list[DIMS]:
    """Identify how the batch dims of input map to the batch dimensions of the output_rvs.

    Example:
    -------
    In the example below `idx` has two batch dimensions (indexed 0, 1 from left to right).
    The two uncommented dependent variables each have 2 batch dimensions where each entry
    results from a mapping of a single entry from one of these batch dimensions.

    This mapping is transposed in the case of the first dependent variable, and shows up in
    the same order for the second dependent variable. Each of the variables as a further
    batch dimension encoded as `None`.

    The commented out third dependent variable combines information from the batch dimensions
    of `idx` via the `sum` operation. A `ValueError` would be raised if we requested the
    connection of batch dims.

    .. code-block:: python
        import pymc as pm

        idx = pm.Bernoulli.dist(shape=(3, 2))
        dep1 = pm.Normal.dist(mu=idx.T[..., None] * 2, shape=(3, 2, 5))
        dep2 = pm.Normal.dist(mu=idx * 2, shape=(7, 2, 3))
        # dep3 = pm.Normal.dist(mu=idx.sum())  # Would raise if requested

        print(subgraph_batch_dim_connection(idx, [], [dep1, dep2]))
        # [(1, 0, None), (None, 0, 1)]

    Returns:
    -------
    list of tuples
        Each tuple corresponds to the batch dimensions of the output_rv in the order they are found in the output.
        None is used to indicate a batch dimension that is not mapped from the input.

    Raises:
    ------
    ValueError
        If input batch dimensions are mixed in the graph leading to output_vars.

    NotImplementedError
        If variable related to marginalized batch_dims is used in an operation that is not yet supported
    """
    var_dims = {input_var: tuple(range(input_var.type.ndim))}
    var_dims = _subgraph_batch_dim_connection(var_dims, [input_var], output_vars)
    ret = []
    for output_var in output_vars:
        output_dims = var_dims.get(output_var, (None,) * output_var.type.ndim)
        assert len(output_dims) == output_var.type.ndim
        ret.append(output_dims)
    return ret
